#!/usr/bin/env python
'''Testing for trends.cgm.py

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import glob
from mock import call, patch
import numpy as np
import numpy.testing as npt
import os
import pdb
import pytest
import unittest

import galaxy_dive.trends.cgm as cgm_trends

########################################################################
########################################################################

class TestCGMTrends( unittest.TestCase ):

    def test_cooling_time( self ):
        '''Just test that we're reading the data correctly.
        '''

        actual = cgm_trends.cooling_time(
            0.70057054,
            1.45,
            'm12m',
            'core',
        )
        # From manually looking at the data
        expected = 0.0006051895628244133

        npt.assert_allclose( expected, actual )
            
    ########################################################################

    def test_cooling_time_array( self ):

        actual = cgm_trends.cooling_time(
            np.array([
                [ 4.015689846912088, 2.8048464526136545, ],
                [ 7.613341094783783, 5.317704697404021, ],
            ]),
            np.array([
                [ 0., 0.5, ],
                [ 0., 0.5, ],
            ]),
            'm12m',
            'core',
        )
        # From manually looking at the data
        expected = np.array([
            [1.00620064e-04, 4.65006698e-05],
            [1.35994157e-04, 4.36442663e-05],
        ])

        npt.assert_allclose( expected, actual )
        
        

    
    




#!/usr/bin/env python
'''Wrapper for hdf5 objects opened with h5py. Serves various needs not inherent in h5py.'''

import copy
import h5py
import numpy as np
import os
import pandas as pd
import six
import shutil

import galaxy_dive.utils.utilities as utilities

########################################################################
__author__ = 'Zachary Hafen'

__maintainer__ = 'Zachary Hafen'
__email__ = 'zachary.h.hafen@gmail.com'
__status__ = 'Beta'

########################################################################

default = object()

########################################################################

class HDF5Wrapper(object):

    def __init__(self, filename):

        self.filename = filename

    ########################################################################

    def save_data(self, data, index_key=None):
        '''Save data at self.filename.

        Args:
            data (dict): The data to be saved.
        '''

        self.f = h5py.File(self.filename, 'a')

        for key in data:

            # If it's a group
            if type(data[key]) == dict:
                self.f.create_group(key)
                for key2 in data[key].keys():
                    if key2 != 'attrs':
                        full_data_key = '{}/{}'.format(key, key2)
                        self.f.create_dataset(full_data_key, data=data[key][key2])

                    # Include any attributes
                    else:
                        for attr_key in data[key]['attrs'].keys():
                            self.f[key].attrs[attr_key] = data[key]['attrs'][attr_key]

            else:
                try:
                    self.f.create_dataset(key, data=data[key])

                # Handle the case where there's an empty dataset
                except TypeError:
                    if data[key][0] == {}:
                        continue
                    else:

                        # Workaround for hdf5 and np.arrays
                        arr = np.array(
                            data[key],
                            dtype=h5py.special_dtype( vlen=six.text_type ),
                        )
                        self.f.create_dataset( key, data=arr )

        # Catalog which key 'Indexes' the dataset, if any.
        if index_key is not None:
            self.f.attrs['index_key'] = index_key

        self.f.close()

    ########################################################################

    def insert_data(self, new_data, index_key=None, insert_columns=False):
        '''Insert data. Does this by making whole new copy, appending to it,
        sorting that by the index, and saving the sorted version.

        Args:
            new_data (dict):
                The new data to be sorted.

            index_key (str):
                What key to sort the data by. Defaults to None, which will use
                the key inherent in the data set, and fail otherwise.

            insert_columns (bool):
                If True, the data in new_data should be included as new columns.
        '''

        try:
            self.f = h5py.File(self.filename, 'r')
        except IOError:
            self.save_data(new_data, index_key)
            return 0

        # Make sure we use a consistent index.
        if index_key is not None:
            if 'index_key' in self.f.attrs.keys():
                assert index_key == self.f.attrs['index_key']
        else:
            if 'index_key' in self.f.attrs.keys():
                index_key = self.f.attrs['index_key']
            else:
                raise Exception('No index_key specified at any point.')

        # Handle an empty data set
        if len(self.f.keys()) == 0:
            self.f.close()
            self.save_data(new_data, index_key)
            return 0

        # Load the data into lists
        data = {}
        for key in self.f.keys():
            data[key] = self.f[key][...].tolist()

        # Append the new data
        if not insert_columns:
            # Case where we have rows.
            try:
                for key in new_data:
                    data[key].append(new_data[key])

            # Deal with single floats
            except AttributeError:

                for key in data:
                    data[key] = [ data[key], ]

                for key in new_data:
                    data[key].append(new_data[key])

        # Columns case (using pandas)
        else:
            # Set up DataFrames
            df = pd.DataFrame( data )
            df_new = pd.DataFrame( new_data )

            # Merge dataframes and get result out
            merged_df = pd.merge( df, df_new, on=index_key )
            data = merged_df.to_dict( 'list' )

        # Sort the data
        data = self.sort_dictionary(data, index_key)

        # Only include unique values, according to the index
        unique_index_data, unique_inds = np.unique(data[index_key], True)
        for key in data:
            data[key] = np.array(data[key])[unique_inds]

        # Delete the old data.
        self.f.close()
        self.clear_hdf5_file()

        # Store the data again
        self.save_data(data, index_key=index_key)

        return 0

    ########################################################################

    def sort_dictionary(self, data, index_key):
        '''Sort a dictionary.

        Args:
            data (dict): Dictionary to sort.
            index_key (str): Key to sort by.
        '''

        # Sort the data
        sorted_inds = np.argsort(data[index_key])
        for key in data:
            data[key] = np.array(data[key])[sorted_inds]

        return data

    ########################################################################

    def subsample_hdf5_file(self, subsamples, particle_data=False ):
        '''Create a dictionary containing subsamples of the data.

        Args:
            subsamples (int): The number of subsamples per axis, for a total of subsamples^n_dims subsamples.
            particle_data (bool or int): Whether or not the data contains particle-like data.
                If not False, then it's the number of files the snapshot is spread across
        '''

        def subsample_dataset(dataset):

            subsampled_data = {}

            if len( dataset ) == 0:
                return subsampled_data
            else:

                # Get the first value for easy reference
                first_value = list( dataset.values() )[0]

                # Get the shape of the subsample
                n_dims = first_value[...].ndim
                subsample_shape = tuple([subsamples for i in range(n_dims)])
                n_subsamples = np.product(subsample_shape)


                # Subsample indices
                if particle_data:
                    n_particles = max( first_value.shape )
                    subsample_inds = np.random.randint(0, n_particles, subsamples)
                else:
                    subsample_inds = np.random.randint(0, first_value.size, n_subsamples)

                # Do the subsampling
                for key in dataset:

                    # Subsample when using particle data
                    if particle_data:

                        n_additional_dims = dataset[key].size // n_particles

                        if n_additional_dims > 1:

                            # Subsample the data
                            subsampled_data_for_key = []
                            for i in range(n_additional_dims):
                                subsampled_data_for_key.append( dataset[key][...][:,i][ subsample_inds ] )

                            # Format the returned data
                            subsampled_data[key] = np.array( subsampled_data_for_key ).transpose()

                        else:

                            subsampled_data[key] = dataset[key][...][ subsample_inds ]

                    # Standard scenario
                    else:
                        values = dataset[key][...].flatten()[subsample_inds]
                        subsampled_data[key] = np.reshape(values, subsample_shape)

                return subsampled_data

        f = h5py.File(self.filename, 'r')

        subsampled_data = {}

        # Subsample differently if we have a number of groups
        first_value = list( f.values() )[0]
        if type(first_value) == h5py._hl.group.Group:
            for key in f.keys():
                subsampled_data[key] = subsample_dataset( f[key] )

                # Include attributes
                group_attrs = {}
                for attr_key in f[key].attrs.keys():

                    # When copying particle data, manually change some of the attrs.
                    if attr_key == 'NumPart_ThisFile':
                        # Don't overwrite when 0 particles
                        nonzero_numpart = f[key].attrs[attr_key] != 0
                        group_attrs[attr_key] = np.where(nonzero_numpart, subsamples, 0)

                    elif attr_key == 'NumPart_Total':
                        # Don't overwrite when 0 particles
                        nonzero_numpart = f[key].attrs[attr_key] != 0
                        group_attrs[attr_key] = np.where(nonzero_numpart, particle_data*subsamples, 0)

                    elif attr_key == 'NumFilesPerSnapshot':
                        group_attrs[attr_key] = particle_data


                    # Standard case
                    else:
                        group_attrs[attr_key] = f[key].attrs[attr_key]

                subsampled_data[key]['attrs'] = group_attrs


        # Standard subsampling
        else:
            subsampled_data = subsample_dataset( f )


        f.close()

        return subsampled_data

    ########################################################################

    def copy_hdf5_file(self, copy_filename, subsamples=False, particle_data=False):
        '''Copy the hdf5 file into a new file. May not copy all attributes successfully?

        Args:
            copy_filename (str): Where the file should be copied to.
            subsamples (int): If chosen, number of subsamples per axis. Defaults to no subsampling.
        '''

        # Open files
        f = h5py.File(self.filename, 'r')

        # Copy the contents
        if not subsamples:

            g = h5py.File(copy_filename, 'a')

            for key in f:
                f.copy(key, g)

            for attr_key in f.attrs:
                g.attrs[attr_key] = f.attrs[attr_key]

        elif subsamples:

            subsampled_data = self.subsample_hdf5_file( subsamples, particle_data=particle_data )

            # Save the data
            g_h5_wrapper = HDF5Wrapper( copy_filename )
            g_h5_wrapper.save_data( subsampled_data )

            g = h5py.File(copy_filename, 'a')

            for attr_key in f.attrs:
                g.attrs[attr_key] = f.attrs[attr_key]

        f.close()

        return g

    ########################################################################

    def clear_hdf5_file(self):
        '''WARNING: Clears the entire hdf5 file of any useful data.'''

        f = h5py.File(self.filename, 'w')
        f.close()

########################################################################

def copy_snapshot( sdir, snum, copy_dir, subsamples=False, redistribute=False, n_files=default ):
    '''Copy a gadget/gizmo snapshot, and subsample it if you choose to.

    Args:
        sdir (str) :
            Snapshot directory to copy from.

        snum (int) :
            Which snapshot to copy over.

        copy_dir (str) :
            Where you want to save the copied snapshot.

        subsamples (bool or int) :
            If not False, the number of particles you will have in your subsampled snapshot,
            per file part.

        redistribute (bool) :
            If True, redistribute the particles evenly among the snapshots.

        n_files (str or int) :
            If not default, the number of pieces you want your copied snapshot broken into
            (must be less than the number of pieces in the original snapshot if subsampling).
            In general, should really only be used when subsampling or redistributing.
    '''

    snapdir = os.path.join( sdir, 'snapdir_{:03}'.format( snum ) )
    copy_snapdir = os.path.join( copy_dir, 'snapdir_{:03}'.format( snum ) )

    # Make sure the path exists
    utilities.make_dir( copy_snapdir )

    assert not (subsamples and redistribute), "Cannot both subsample and redistribute"

    n_files_orig = len( os.listdir( snapdir ) )
    if n_files is default:
        n_files = n_files_orig

    print( 'Starting copying...' )

    if redistribute:

        orig_file_basename = 'snapshot_{:03}.0.hdf5'.format( snum, 0 )
        orig_filename = os.path.join( snapdir, orig_file_basename )

        print( 'Loading all data...' )
        data = {}
        for i in range( n_files_orig ):

            print('  Loading file {}...'.format( i ))

            file_basename = 'snapshot_{:03}.{}.hdf5'.format( snum, i )
            filename = os.path.join( snapdir, file_basename )

            with h5py.File( filename, 'r' ) as f:

                ptypes = copy.copy( list( f.keys() ) )
                del ptypes[0]

                assert not ( 'Header' in ptypes ), "Deleted wrong key..."

                for ptype in ptypes:

                    if i == 0:
                        data[ptype] = {}

                    pdata = f[ptype]
                    for key in pdata.keys():

                        if i == 0:
                            data[ptype][key] = []

                        data[ptype][key].append( pdata[key][...] )

        print( 'Concatenating and splitting data...' )
        for ptype in data.keys():
            print( '  Concatenating {} data'.format( ptype ) )
            for key in data[ptype].keys():

                # Pull into a single snapshot
                data[ptype][key] = np.concatenate( data[ptype][key], axis=0 )
                data[ptype][key] = np.array_split( data[ptype][key], n_files, axis=0 )

        print( 'Storing data...' )
        for i in range( n_files ):

            print('  Copying {} of {}'.format( i+1, n_files ))

            file_basename = 'snapshot_{:03}.{}.hdf5'.format( snum, i )
            copy_filename = os.path.join( copy_snapdir, file_basename )

            shutil.copyfile( filename, copy_filename )

            print( '  Storing {} of {}'.format( i+1, n_files ) )

            with h5py.File( copy_filename, 'a' ) as f:

                n_particles_file = [0,]*len( f['Header'].attrs['NumPart_ThisFile'] )

                for ptype in ptypes:

                    n_particles_file[int(ptype[-1])] = data[ptype]['Masses'][i].size
                    for key in data[ptype].keys():

                        del f[ptype][key]
                        f[ptype][key] = data[ptype][key][i]

                    del f['Header'].attrs['NumPart_ThisFile']
                    f['Header'].attrs['NumPart_ThisFile'] = np.array( n_particles_file ).astype( np.int32 )

                del f['Header'].attrs['NumFilesPerSnapshot']
                f['Header'].attrs['NumFilesPerSnapshot'] = np.int32( n_files )

    else:
        for i in range( n_files ):

            print('  Copying {} of {}'.format( i+1, n_files ))

            file_basename = 'snapshot_{:03}.{}.hdf5'.format( snum, i )
            filename = os.path.join( snapdir, file_basename )
            copy_filename = os.path.join( copy_snapdir, file_basename )

            # Actually make the copy
            h5_wrapper = HDF5Wrapper( filename )
            h5_wrapper.copy_hdf5_file( copy_filename, subsamples=subsamples, particle_data=n_files, )

    print('Done!')




#!/usr/bin/env python
'''Tools for reading AHF output files.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import glob
import numba
import numpy as np
import os
import pandas as pd
import unyt

import galaxy_dive.read_data.metafile as read_metafile

########################################################################

default = object()

########################################################################
########################################################################

class AHFReader( object ):
    '''Read AHF data.
    Note! All positions are in comoving coordinates, and everything has 1/h's sprinkled throughout.
    '''

    def __init__( self, data_dir ):
        '''Initializes.

        Args:
            data_dir (str): Simulation directory to load the AHF data from.
        '''

        self.data_dir = data_dir

        # Common global names, and what columns those correspond to in the
        # data
        self.general_use_data_names = {
            'X' : 'Xc',
            'Y' : 'Yc',
            'Z' : 'Zc',
            'Vx' : 'VXc',
            'Vy' : 'VYc',
            'Vz' : 'VZc',
        }

        # Data units
        self.units = {
            'Mvir' : unyt.Msun,
            'Xc' : unyt.kpc,
            'Yc' : unyt.kpc,
            'Zc' : unyt.kpc,
            'VXc' : unyt.km / unyt.s,
            'VYc' : unyt.km / unyt.s,
            'VZc' : unyt.km / unyt.s,
            'Rvir' : unyt.kpc,
            'Rstar0.5' : unyt.kpc,
            'Rstar0.9' : unyt.kpc,
        }

    ########################################################################
    # Load Data
    ########################################################################

    def get_mtree_halos( self, index=None, tag=None, adjust_default_labels=default, ):
        '''Get halo files (e.g. halo_00000.dat) in a dictionary of pandas DataFrames.

        Args:
            index (str or int) :
                What type of index to use. Defaults to None, which raises an exception. You *must* choose
                an index, to avoid easy mistakes. Options are...
                'snum' : Indexes by snapshot number, starting at 600 and counting down. Only viable with snapshot steps of 1!!
                                  Identical to passing the integer 600. (See below.)
                'range' : Index by an increasing range.
                int : If an integer, then this integer should be the final snapshot number for the simulation.
                            In this case, indexes by snapshot number, starting at the final snapshot number and counting down.
                            Only viable with snapshot steps of 1!!

            tag (str) :
                Additional identifying tag for the files, e.g. 'smooth', means this function will look for
                'halo_00000_smooth.dat', etc.

            adjust_default_label (bool) :
                Whether or not to adjust the column headings, etc. Defaults to True if tag is None.

        Modifies:
            self.mtree_halos (dict of pd.DataFrames) :
                DataFrames containing the requested data. The key for a given dataframe
                is that dataframe's Merger Tree Halo ID

            self.mtree_halo_filepaths (dict of strs) :
                What files the data was loaded from.

            self.index (str) :
                The users value for the index.

            self.tag (str) :
                What tag was used for the files.
        '''

        if adjust_default_labels is default:
            if tag is None:
                adjust_default_labels = True
            else:
                adjust_default_labels = False

        def get_halo_filepaths( unexpanded_filename ):
            '''Function for getting a list of filepaths'''
            filepath_unexpanded = os.path.join( self.data_dir, unexpanded_filename )
            halo_filepaths = glob.glob( filepath_unexpanded )
            return set( halo_filepaths )

        # Get the filename to search for
        if tag is not None:
            ahf_filename = 'halo_*_{}.dat'.format( tag )
            halo_filepaths = get_halo_filepaths( ahf_filename )

        else:
            ahf_filename = 'halo_*.dat'
            halo_filepaths = get_halo_filepaths( ahf_filename )

            # Find all files that are modified.
            ahf_modified_filename = 'halo_*_*.dat'
            halo_modified_filepaths = get_halo_filepaths( ahf_modified_filename )

            # Remove all the modified filepaths from the search list.
            halo_filepaths -= halo_modified_filepaths

        # Raise an exception if there are no files to load
        if len( halo_filepaths ) == 0:
            raise NameError( 'No files to load in {}'.format( self.data_dir ) )

        # Set up the data storage
        self.mtree_halos = {}
        self.mtree_halo_filepaths = {}

        # Loop over each file and load it
        for halo_filepath in halo_filepaths:

            # Load the data
            mtree_halo = pd.read_csv( halo_filepath, sep='\t', )

            # Extra tweaking to read the default AHF file format
            if adjust_default_labels:
                # Delete a column that shows up as a result of formatting
                del mtree_halo[ 'Unnamed: 93' ]

                # Remove the annoying parenthesis at the end of each label.
                mtree_halo.columns = [ label.split( '(' )[0] for label in list( mtree_halo ) ]

                # Remove the pound sign in front of the first column's name
                mtree_halo = mtree_halo.rename( columns = {'#redshift':'redshift', ' ID':'ID'} )

            # Get a good key
            base_filename = os.path.basename( halo_filepath )
            halo_num_str = base_filename[5:]
            if tag is not None:
                halo_num_str = halo_num_str.split( '_' )[0]
            halo_num = int( halo_num_str.split( '.' )[0] )

            if index == 'range':
                pass
            elif (index == 'snum') or isinstance( index, int ):
                if index == 'snum':
                    final_snapshot_number = 600
                else:
                    final_snapshot_number = index
                # Set the index, assuming we have steps of one snapshot
                n_rows = mtree_halo.shape[0]
                mtree_halo['snum'] = range( final_snapshot_number, final_snapshot_number - n_rows, -1)
                mtree_halo = mtree_halo.set_index( 'snum', )
            elif 'snum' in mtree_halo.columns:
                mtree_halo = mtree_halo.set_index( 'snum', )
            else:
                raise Exception( "index type not selected" )

            # Store the data
            self.mtree_halos[ halo_num ] = mtree_halo
            self.mtree_halo_filepaths[ halo_num ] = halo_filepath

        # Save the index and tag as an attribute
        self.index = index
        self.tag = tag

    ########################################################################

    def get_halos( self, snum, force_reload=False ):
        '''Get *.AHF_halos file for a particular snapshot.

        Args:
            snum (int): Snapshot number to load.

            force_reload (bool): Force reloading, even if there's already an halos file loaded.

        Modifies:
            self.halos (pd.DataFrame): Dataframe containing the requested data.
        '''

        if hasattr( self, 'halos' ):
            if (self.halos_snum == snum) and not force_reload:
                return

        # Load the data
        self.halos_path = self.get_filepath( snum, 'AHF_halos' )
        self.halos = pd.read_csv( self.halos_path, sep='\t', index_col=0 )

        # Delete a column that shows up as a result of formatting
        del self.halos[ 'Unnamed: 92' ]

        # Remove the annoying parenthesis at the end of each label.
        self.halos.columns = [ label.split('(')[0] for label in list( self.halos ) ]

        # Rename the index to a more suitable name, without the '#' and the (1)
        self.halos.index.names = ['ID']

        # Save the snapshot number of the ahf halos file.
        self.halos_snum = snum

        # Note that we haven't added the additional halos data yet.
        self.halos_added = False

    ########################################################################

    def get_halos_add( self, snum, force_reload=False ):
        '''Get *.AHF_halos_add file for a particular snapshot.

        Args:
            snum (int): Snapshot number to load.

            force_reload (bool): Force reloading, even if there's already an halos file loaded.

        Modifies:
            self.halos_add (pd.DataFrame): Dataframe containing the requested data.
        '''

        if self.halos_added:
            if (self.halos_snum == snum) and not force_reload:
                return

        # Load the data
        self.halos_add_path = self.get_filepath( snum, 'AHF_halos_add' )
        halos_add = pd.read_csv( self.halos_add_path, sep='\t', index_col=0 )

        if not hasattr( self, 'halos' ) :
            self.get_halos( snum )
        if self.halos_snum != snum:
            self.get_halos( snum )

        self.halos = pd.concat( [ self.halos, halos_add ], axis=1 )

        self.halos_added = True

    ########################################################################

    def get_mtree_idx( self, snum ):
        '''Get *.AHF_mtree_idx file for a particular snapshot.

        Args:
            snum (int): Snapshot number to load.

        Modifies:
            self.ahf_mtree_idx (pd.DataFrame): Dataframe containing the requested data.
        '''

        # If the data's already loaded, don't load it again.
        if hasattr( self, 'ahf_mtree_idx' ):
            return self.ahf_mtree_idx

        # Load the data
        ahf_mtree_idx_path = self.get_filepath( snum, 'AHF_mtree_idx' )
        self.ahf_mtree_idx = pd.read_csv( ahf_mtree_idx_path, delim_whitespace=True, names=['HaloID(1)', 'HaloID(2)'], skiprows=1  )

    ########################################################################

    def get_profiles( self, snum ):
        '''Get *.AHF_mtree_idx file for a particular snapshot.

        Args:
            snum (int): Snapshot number to load.

        Modifies:
            self.ahf_mtree_idx (pd.DataFrame): Dataframe containing the requested data.
        '''

        # If the data's already loaded, don't load it again.
        if hasattr( self, 'profiles' ):
            return self.profiles

        # Load the data
        profiles_path = self.get_filepath( snum, 'AHF_profiles' )
        self.profiles = pd.read_csv(
            profiles_path,
            sep = '\t', 
        )

        # Delete a column that shows up as a result of formatting
        del self.profiles['Unnamed: 27']

        # Remove the annoying parenthesis at the end of each label.
        self.profiles.columns = [ label.split('(')[0] for label in list( self.profiles ) ]

        # Remove the # at the beginning of the first column
        self.profiles.rename( columns={'#r': 'r'}, inplace=True )

        # Get the mapping for where the halo IDs are in the profile data
        @numba.jit( 'i8[:](f8[:])', nopython=True )
        def profile_mapping( r ):
            
            ind_mapping = [ 0, ]
            for i, r_i in enumerate( r ):

                # Skip first loop
                if i==0:
                    continue

                # Record when the values start over
                if r_i <  r[i-1]:
                    ind_mapping.append( i )

                # Last index
                if i == r.size - 1:
                    ind_mapping.append( i+1 )

            return np.array( ind_mapping )

        # Run the function
        self.profile_id_mapping = profile_mapping(
            np.abs( self.profiles['r'].values )
        )

    ########################################################################
    # Utilities
    ########################################################################

    def get_filepath( self, snum, ahf_file_type ):
        '''Get the filepath for a specified type of AHF file.

        Args:
            snum (int): Snapshot number to load.
            ahf_file_type (str): Can be AHF_halos or AHF_mtree_idx.

        Returns:
            ahf_filepath (str): The filepath to the specified file.
        '''

        # Load the data
        ahf_filename = 'snap{:03d}Rpep..z*.*.{}'.format( snum, ahf_file_type )
        ahf_filepath_unexpanded = os.path.join( self.data_dir, ahf_filename )
        possible_filepaths = glob.glob( ahf_filepath_unexpanded )
        if len( possible_filepaths ) == 0:
            raise NameError( 'No files to load for snum {} in {}'.format( snum, self.data_dir ) )
        elif len( possible_filepaths ) > 1:
            raise Exception( 'Multiple possible *.{} files to load'.format( ahf_file_type ) )
        ahf_filepath = possible_filepaths[0]

        return ahf_filepath

    ########################################################################

    def check_files_exist( self, snum_start, snum_end, snum_step, file_str='AHF_halos' ):
        '''Check that AHF was successfully run for all files in the given range.

        Args:
            snum_start (int) : Starting snapshot.
            snum_end (int) : Ending snapshot.
            snum_step (int) : Step between snapshots.

        Returns:
            files_exist (bool) : If True, all files in the given range exist.
        '''

        snums = range( snum_start, snum_end + 1, snum_step )

        missing_snums = []
        for snum in snums:

            file_to_check = 'snap{:03d}Rpep*{}'.format( snum, file_str )
            filepath_to_check = os.path.join( self.data_dir, file_to_check )
            files = glob.glob( filepath_to_check )

            missing_ahf_files_at_this_snapshot = len( files ) == 0
            if missing_ahf_files_at_this_snapshot:
                missing_snums.append( snum )

        if len( missing_snums ) > 0:

            print( 'Missing snums:' )
            for missing_snum in missing_snums:
                print( '{},'.format( missing_snum ) )

            return False

        return True

    ########################################################################
    # Get Specific Data Arrays
    ########################################################################

    def get_mtree_halo_quantity( self, quantity, indice, index=None, tag=None ):
        '''Get a desired quantity for all merger tree halos at a particular snapshot.

        Args:
            quantity (str): mtree_halo key to load in the dataset
            indice (int): Indice of the quantity to load, as indicated by the index.
            index (str or int) : What type of index to use. Defaults to None, which raises an exception.
                You *must* choose an index, to avoid easy mistakes. See get_mtree_halos() for a full description.
            tag (str) : Additional identifying tag for the files, e.g. 'smooth', means this function will look for
                'halo_00000_smooth.dat', etc.

        Returns:
            mtree_halo_quantity (np.array): The ith index is the requested quantity for ith MT halo.
        '''

        # Load the data if it's not already loaded.
        if not hasattr( self, 'mtree_halos' ):
            self.get_mtree_halos( index=index, tag=tag )
        else:
            assert index == self.index
            assert tag == self.tag

        mtree_halo_quantity = []
        for halo_id in sorted( self.mtree_halos.keys() ):

            try:
                mtree_halo_quantity.append( self.mtree_halos[ halo_id ][ quantity ][ indice ] )

            # When we're past the point of galaxies being identified (i.e. high redshift, galaxies aren't formed yet),
            # set the values by hand.
            except KeyError:

                if quantity == 'Mvir':
                    mtree_halo_quantity.append( 0. )

                else:
                    raise Exception( "Value of {} not specified before galaxies form".format( quantity ) )

        return np.array( mtree_halo_quantity )

    ########################################################################

    def get_pos_or_vel( self, pos_or_vel, halo_id, inds, type_of_halo_id='merger_tree' ):
        '''Get the position or velocity of a mt halo (three dimensional).

        Args:
            pos_or_vel (str): Get position ('pos') or velocity ('vel').
            halo_id (int): Merger tree halo ID for the position or velocity you want.
            inds (int or np.array of ints): Indices you want the position or velocity for.
              If type_of_halo_id == 'merger_tree', uses same index as mtree_halos.
              Elif type_of_halo_id == 'halos', can only be a single int,
              which should be the snapshot number.
            type_of_halo_id (str): 'merger_tree' if the halo id is a merger tree halo id.
                                                          'halos' if the halo id is a *.AHF_halos halo id.

        Returns:
            p_or_v ( [len(inds), 3] np.array ): Position or velocity for the specified inds.
        '''

        # Choose the indices we'll access the data through
        if pos_or_vel == 'pos':
            keys = [ 'Xc', 'Yc', 'Zc' ]
        elif pos_or_vel == 'vel':
            keys = [ 'VXc', 'VYc', 'VZc' ]
        else:
            raise Exception( 'Unrecognized pos_or_vel, {}'.format( pos_or_vel ) )

        # Get the ahf_halo data, if requested.
        if type_of_halo_id == 'halos':
            self.get_halos( inds )

        # Get the data.
        p_or_v = []
        for key in keys:

            # Get the part
            if type_of_halo_id == 'merger_tree':
                p_or_v_part = self.mtree_halos[ halo_id ][ key ][ inds ]
            elif type_of_halo_id == 'halos':
                p_or_v_part = self.halos[ key ][ halo_id ]
            else:
                raise Exception( 'Unrecognized type_of_halo_id, {}'.format( type_of_halo_id ) )

            p_or_v.append( p_or_v_part )

        # Finish formatting.
        p_or_v = np.array( p_or_v ).transpose()

        return p_or_v




#!/usr/bin/env python
'''General utilities

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import collections
import copy
from contextlib import contextmanager
import errno
from functools import wraps
import h5py
import importlib
import inspect
import itertools
import numpy as np
import os
import shutil
import six
import subprocess
import sys
import time
# Works for py2 and py3
try:
    from StringIO import StringIO
except ImportError:
        from io import StringIO

########################################################################

def merge_two_dicts( dict_a, dict_b ):
    '''Merges two dictionaries into a shallow copy.

    Args:
        dict_a (dict) : First dictionary to merge.
        dict_b (dict) : Second dictionary to merge.

    Returns:
        merged_dict (dict) :
            Dictionary including elements from both.
            dict_a's entries take priority over dict_b.
    '''

    merged_dict = dict_b.copy()
    merged_dict.update( dict_a )

    return merged_dict

########################################################################

def dict_from_defaults_and_variations( defaults, variations ):
    '''Create a dictionary of dictionaries from a default dictionary and variations on it.

    Args:
        defaults (dict) : Default dictionary. What each individual dictionary should default to.
        variations (dict of dicts) : Each dictionary contains what should be different. The key for each dictionary should
            be a label for it.

    Returns:
        result (dict of dicts) : The results is basically variations, where each child dict is merged with defaults.
    '''

    if len( defaults ) == 0:
        return variations

    result = {}
    for key in variations.keys():

        defaults_copy = defaults.copy()

        defaults_copy.update( variations[key] )

        result[key] = defaults_copy

    return result

########################################################################

def deepgetattr( obj, attr ):
    '''Recurses through an attribute chain to get the ultimate value.
    Credit to http://pingfive.typepad.com/blog/2010/04/deep-getattr-python-function.html

    Args:
        obj (object) : Object for which to get the attribute.
        attr (str) : Attribute to get. Can be nested, e.g. obj.foo.bar.dog

    Returns:
        result (attr object) : The requested attribute.
    '''

    return reduce( getattr, attr.split('.'), obj )

########################################################################

def arrays_to_set( *args ):
    '''Convert arr1, arr2, arr3, ... to a set, where a single enrty in the set is ( arr1[j], arr2[j], ... )

    Args:
        n arrays ( multiple np.ndarrays ) : The arrays to convert.

    Returns:
        result ( set ) : The converted set.
    '''

    return set( zip( *args ) )

def set_to_arrays( set_to_convert ):
    '''Convert a set into multiple arrays.

    Args:
        result ( set ) : The set to convert

    Returns:
        n arrays ( multiple np.ndarrays ) : The converted arrays
    '''

    return np.array( list( set_to_convert ) ).transpose()

########################################################################

@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err

########################################################################

def chunk_list( l, n ):
    '''Breaks a list l into n chunks, as equally as possible.

    Args:
        l (list) : List to break into chunks.
        n (int) : Number of chunks to break the list into.

    Returns:
        chunked_l ( list of lists ) : The list broken into chunks.
    '''

    start_ind = 0
    end_ind = 0
    remainder_to_distribute = len( l ) % n

    chunked_l = []
    for i in range( n ):

        # l/n may not always be an int, so forcing it to be one as before.
        # But the fact that it isn't always an int seems like a bigger problem.
        end_ind += int(len( l )/n)

        if remainder_to_distribute > 0:
            end_ind += 1
            remainder_to_distribute -= 1

        chunk = l[ start_ind : end_ind ]

        chunked_l.append( chunk )

        start_ind = end_ind

    return chunked_l

########################################################################

def get_instance_source_dir( instance, instance_type='class' ):
    '''Get the directory containing the source code of an instance (class or module).

    instance (object) : Instance to get the code for.
    instance_type (str) : 'class' if the instance is a class, 'module' if the instance is a module
    '''

    if instance_type == 'class':
        inspection_object =  instance.__class__
    elif instance_type == 'module':
        inspection_object = instance
    else:
        raise KeyError( "Unrecognized instance_type, {}".format( instance_type ) )

    sourcefile = inspect.getabsfile( inspection_object )

    sourcedir = os.path.dirname( sourcefile )

    return sourcedir

########################################################################

def get_code_version( instance, instance_type='class' ):
    '''Get the current version (git tag/commit) of the parent code for a class or module

    instance (object) : Instance to get the code for.
    instance_type (str) : 'class' if the instance is a class, 'module' if the instance is a module
    '''

    cwd = os.getcwd()

    # Change to the directory of the instance
    instancedir = get_instance_source_dir( instance, instance_type=instance_type )
    os.chdir( instancedir )

    # Get the code version
    try:
        code_version = subprocess.check_output( [ 'git', 'describe', '--always' ] )

    # If we fail, don't break the code.
    except:
        code_version = 'failed to get version'

    # Change back
    finally:
        os.chdir( cwd )

        return code_version

########################################################################

def print_timer( timer_string='Time taken:' ):
    '''Decorator to time a function.

    Args:
        timer_string (str, optional) : Printed before printing the time.
    '''

    def _print_timer( func ):

        @wraps( func )
        def wrapped_func( *args, **kwargs ):

            time_start = time.time()

            result = func( *args, **kwargs )

            time_end = time.time()

            print_string = '{} {:.3g} seconds'.format( timer_string, time_end - time_start )
            print( print_string )

            return result

        return wrapped_func

    return _print_timer

########################################################################

def store_parameters( constructor ):
    '''Decorator for automatically storing arguments passed to a constructor.
    I.e. any args passed to constructor via test_object = TestObject( *args, **kwargs )
    will be stored in test_object, e.g. test_object.args

    Args:
        constructor (function) : Constructor to wrap.
    '''

    @wraps( constructor )
    def wrapped_constructor( self, *args, **kwargs ):

        parameters_to_store = inspect.getcallargs( constructor, self, *args, **kwargs )

        # Make sure we don't accidentally try to save the self argument
        del parameters_to_store['self']

        for parameter in parameters_to_store.keys():
            setattr( self, parameter, parameters_to_store[parameter] )

        self.stored_parameters = list( parameters_to_store.keys() )

        result = constructor( self, *args, **kwargs )

    return wrapped_constructor

########################################################################

def save_parameters( instance, f ):
    '''Save parameters to a hdf5 file.

    Args:
        instance (object) :
            Instance that has the attributes that are parameters.
            Typically stored in instance.stored_parameters.

        f (open h5py file object) :
            File to save the parameters to.

    Returns:
        param_grp (h5py group) :
            Group containing the parameters as attributes.
    '''

    param_grp = f.create_group( 'parameters' )
    for parameter_str in instance.stored_parameters:

        parameter = getattr( instance, parameter_str )

        try:
            if parameter is None:
                param_grp.attrs[parameter_str] = 'None'
            else:
                param_grp.attrs[parameter_str] = parameter

        except TypeError:
            try:
                # Workaround for h5py poorly handling np string arrays.
                parameter = np.array(
                    parameter,
                    dtype = h5py.special_dtype( vlen=six.text_type ),
                )
                param_grp.attrs[parameter_str] = parameter
            except TypeError:
                raise TypeError(
                    "Parameter {} = {} failed to save.".format(
                            parameter_str,
                            parameter
                        )
                    )

    return param_grp

########################################################################

def check_and_decode_bytes( x, coding='utf-8' ):
    '''Check if an object is a bytes object, if so decode it into a string.

    Args:
        x (str or bytes):
            Object to check.

        coding (str):
            What coding to use for decoding bytes objects.

    Returns:
        str:
            The original string or the decoded x.
    '''

    if isinstance( x, bytes ):
        x_out = x.decode( coding )
    else:
        x_out = copy.deepcopy( x )

    return x_out

########################################################################

def gen_derived_data_doc(
    object_import_path,
    save_filepath,
    header_file = None,
):
    '''Generate a document that lists the methods for calculating derived data
    quantities.

    Args:
        object_import_path (str):
            Name of the object to document the derived quantities of.

        save_filepath (str):
            Where to save the result.

        header_file (str):
            What to use as the header for the file, that explains what the
            document contains.
            Defaults to galaxy_dive/utils/derived_data_header.rst.
        
    Modifies:
        file at save_filepath (.rst file):
            Saves the output here.
    '''

    # Process the path
    module_path, class_name = os.path.splitext( object_import_path )
    # Remove the period
    class_name = class_name[1:]

    # Get the object out
    module = importlib.import_module( module_path )
    cl = getattr( module, class_name )

    # Setup the file
    if header_file is None:
        header_file = os.path.join(
            os.path.dirname( __file__ ),
            'derived_data_header.rst',
        )
    shutil.copy( header_file, save_filepath )

    # Write the automatically generated part
    with open( save_filepath, 'a' ) as f:

        # Class line
        f.write( '\n.. autoclass:: {}\n'.format( object_import_path ) )
        f.write( '    :show-inheritance:\n\n' )

        # Write out derived data methods
        for attr in dir( cl ):
            if attr[:5] == 'calc_':
                f.write( '    .. automethod:: {}\n'.format( attr ) )

########################################################################

# Make a path to a file

def make_dir( path ):

    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise

    return 0

########################################################################

# Class modelled after unix tee, mainly for the purpose of writing to a log while also printing.

class Tee:
    def write( self, *args, **kwargs ):
        self.out1.write( *args, **kwargs )
        self.out2.write( *args, **kwargs )
    def __init__( self, out1, out2 ):
        self.out1 = out1
        self.out2 = out2

########################################################################

# Function for saving to a file.
def saveToLog( log_save_file ):

    sys.stdout = Tee(open(log_save_file, 'w'), sys.stdout)

    return 0

########################################################################
# Tools for dealing with cosmology-specific data files.
########################################################################

# Get information from a halo file for a particular redshift.

def getHaloDataRedshift(sdir, halo_number, redshift, convert_to_proper=True):
# Values returned are, in order, and in comoving units,
#   redshift, halo_ID, host_ID, xpeak, ypeak, zpeak, Rvir, Mvir, Mgas, Mstar, xCOM, yCOM, zCOM.

    # Get the halo data for all snapshots
    halo_file = getHaloFilename(sdir, halo_number)
    full_halo_data = dataio.tuples_to_arrays(dataio.read_data(halo_file))
    redshift_arr = full_halo_data[0]

    # Account for rounding errors at the edges.
    if redshift < redshift_arr.min():
        print( 'zhh_dataio.getHaloDataRedshift: Bumping redshift from {} to {}'.format(redshift, redshift_arr.min()) )
        redshift = redshift_arr.min()
    elif redshift > redshift_arr.max():
        print( 'zhh_dataio.getHaloDataRedshift: Bumping redshift from {} to {}'.format(redshift, redshift_arr.max()) )
        redshift = redshift_arr.max()

    try:
        # Get single values out of the halo data.
        arr_value_list = []
        for arr in full_halo_data:
            arr_interp = scipy.interpolate.interp1d(redshift_arr[::-1], arr[::-1])
            arr_value = arr_interp(redshift)
            arr_value_list.append(arr_value)
    except ValueError:
        arr_value_list = np.array(full_halo_data).flatten()

    # Convert halo data to proper coordinates (the 1/h factor's still there for many of the coords)
    if convert_to_proper==True:
        comoving_vals = [3, 4, 5, 6, 10, 11, 12]
        for val_index in comoving_vals:
            arr_value_list[val_index] /= (1. + arr_value_list[0])

    return arr_value_list

########################################################################

# Get particle data and useful halo file data for a given data, all in proper units. Combine them into a class.

class PartDataLong(object):

    def __init__(self, sdir, snum, ptype, halo_number=0):
        self.sdir = sdir
        self.snum = snum
        self.ptype = ptype
        self.halo_number = halo_number

        # Get the particles from a snapshot (in proper units)
        self.P = readsnap.readsnap(sdir, snum, ptype, cosmological=1)

        # Get the halo data
        halo_data = getHaloDataRedshift(sdir, halo_number, self.P['redshift'])

        # Get rid of 1/h factors in the halo data
        vals_to_be_converted = range(3, 13)
        for val_index in vals_to_be_converted:
            halo_data[val_index] /= self.P['hubble']

        # Add the halo data to the class.
        self.redshift = halo_data[0]
        self.halo_ID = halo_data[1]
        self.host_ID = halo_data[2]
        self.peak_coords = (halo_data[3], halo_data[4], halo_data[5])
        self.R_vir = halo_data[6]
        self.M_vir = halo_data[7]
        self.M_gas = halo_data[8]
        self.M_star = halo_data[9]
        self.CoM_coords = (halo_data[10], halo_data[11], halo_data[12])

    ########################################################################

    # Change coordinates to center on the main halo.
    def changeCoordsCenter(self):
        self.P['p'] -= np.array(self.peak_coords)

    ########################################################################

    # Calculate the distance from the origin for a given particle.
    def calcRadialDistance(self):

        coords = self.P['p']
        x, y, z = coords.transpose()
        self.P['r'] = np.sqrt(x**2. + y**2. + z**2.)

        return self.P['r']




#!/usr/bin/env python
'''Tools for handling halo data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import copy
import glob
import numpy as np
import os
import pandas as pd

import galaxy_dive.read_data.ahf as read_ahf
import galaxy_dive.read_data.metafile as read_metafile
import galaxy_dive.utils.utilities as utilities
import galaxy_dive.analyze_data.generic_data as generic_data

########################################################################
########################################################################

class HaloData( generic_data.GenericData ):

  @utilities.store_parameters
  def __init__( self, data_dir, tag=None, index=None, mt_kwargs={} ):
    '''Constructor for HaloData

    Args:
      data_dir (str) : Directory storing the data.
      tag (str) : If provided, what is an identifying tag for the halo data?
      index (int) : If provided, what is the final snapshot number for the halo data? Necessary for some AHF data.
      mt_kwargs (dict) : When loading merger tree halo files, additional arguments should be passed here.
    '''

    self.ahf_reader = read_ahf.AHFReader( data_dir )

    key_parser = HaloKeyParser()

    super( HaloData, self ).__init__( key_parser=key_parser )

  ########################################################################
  # Properties
  ########################################################################

  @property
  def mt_halos( self ):
    '''Attribute for accessing merger tree data.
    '''

    if not hasattr( self.ahf_reader, 'mtree_halos' ):

      self.ahf_reader.get_mtree_halos( index=self.index, tag=self.tag, **self.mt_kwargs )

    return self.ahf_reader.mtree_halos

  ########################################################################
  # Data Retrieval
  ########################################################################

  def get_data( self, data_key, snum ):
    '''Get halo data at a specific snapshot.

    Args:
      data_key (str) : What data to get.
      snum (int) : What snapshot to open.

    Returns:
      data (np.ndarray) : Requested data.
    '''

    self.ahf_reader.get_halos( snum )
    self.ahf_reader.get_halos_add( snum )

    return self.ahf_reader.ahf_halos[data_key].values

  ########################################################################

  def get_mt_data( self, data_key, mt_halo_id=0, snums=None, a_power=None, return_values_only=True ):
    '''Get halo data for a specific merger tree.

    Args:
      data_key (str) : What data to get.
      mt_halo_id (int) : What merger tree halo ID to select.
      snums (array-like) : If specified, get the values at these snapshots.
      a_power (float) : If given, multiply the result by the scale factor 1/(1 + redshift) to this power.
      return_values_only (bool) : If True, get rid of pandas data formatting

    Returns:
      mt_data (np.ndarray) : Requested data.
    '''

    mt_data = copy.copy( self.mt_halos[mt_halo_id][data_key] )

    # For converting coordinates
    if a_power is not None:
      mt_data *= ( 1. + self.get_mt_data( 'redshift', mt_halo_id ) )**-a_power

    if snums is not None:
      mt_data = mt_data.loc[snums]

    if return_values_only:
     mt_data = mt_data.values

    return mt_data

  ########################################################################

  def get_selected_data( self, *args, **kwargs ):

    return super( HaloData, self ).get_selected_data( mask_multidim_data=False, *args, **kwargs )

########################################################################
########################################################################

class HaloKeyParser( generic_data.DataKeyParser ):

  def get_radius_key( self, multiplier, length_scale ):
    '''Get a key for Halo data, based on a length scale and a multiple of it.

    Args:
      multiplier (float) :
        multiplier*length_scale defines the radius around the center of the halo(s).

      length_scale (str) :
        multiplier*length_scale defines the radius around the center of the halo(s).

    Returns:
      radius_key (str) :
        Combination of length_scale and multiplier.
    '''

    if np.isclose( multiplier, 1.0 ):
      radius_key = length_scale
    else:
      radius_key = '{}{}'.format( multiplier, length_scale )

    return radius_key

  ########################################################################

  def get_enclosed_mass_key( self, ptype, multiplier, length_scale ):
    '''Get a key for Halo data, corresponding to a data column that records an enclosed mass.

    Args:
      ptype (str) :
        The particle type for the enclosed mass.

      multiplier (float) :
        multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

      length_scale (str) :
        multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

    Returns:
      enclosed_mass_key (str)
    '''

    return 'M{}({})'.format( ptype, self.get_radius_key( multiplier, length_scale ) )

  ########################################################################

  def get_average_quantity_key( self, data_key, ptype, multiplier, length_scale ):
    '''Get a key for Halo data, corresponding to a data column that records the average quantity inside a galaxy.

    Args:
      data_key (str) :
        What the enclosed quantity is.

      ptype (str) :
        The particle type for the enclosed mass.

      multiplier (float) :
        multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

      length_scale (str) :
        multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

    Returns:
      average_quantity_key (str)
    '''

    return '{}{}({})'.format( data_key, ptype, self.get_radius_key( multiplier, length_scale ) )


  ########################################################################

  def get_velocity_at_radius_key( self, velocity_key, multiplier, length_scale ):
    '''Get a key for Halo data, corresponding to a data column that records the velocity at at a specified radius

      velocity_key (str) :
        What velocity to get.

      multiplier (float) :
        multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

      length_scale (str) :
        multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

    Returns:
      velocity_at_radius_key (str)
    '''

    return '{}({})'.format( velocity_key, self.get_radius_key( multiplier, length_scale ) )




'''Testing for astro.py
'''

from mock import patch
import numpy as np
import numpy.testing as npt
import pdb
import unittest

import galaxy_dive.utils.astro as astro
import galaxy_dive.utils.constants as constants

########################################################################

class TestCircularVelocity( unittest.TestCase ):

  ########################################################################

  def test_circular_velocity( self ):

    r_vir = np.array([ 268.00569698,  255.8877719 ,  239.19529116]) 
    m_vir = np.array([  1.07306268e+12,   1.04949145e+12,   1.01265385e+12])

    # What our actual circular velocity is
    result = astro.circular_velocity( r_vir, m_vir )

    # Make sure we have the right dimensions
    assert result.shape == ( 3, )

    # We expect the circular velocity of a 1e12 Msun galaxy to be roughly ~100 km/s
    expected = 100.
    actual = result[0]
    npt.assert_allclose( expected, actual, rtol=0.5 )

  ########################################################################

  def test_hubble_parameter( self ):

    expected = 75.71

    actual = astro.hubble_parameter( 0.16946, .702, 0.272, 0.728 )

    npt.assert_allclose( expected, actual, rtol=1e-4 )

  ########################################################################


  def test_hubble_parameter_1s( self ):

    expected = 75.71*3.24078e-20

    actual = astro.hubble_parameter( 0.16946, .702, 0.272, 0.728, units='1/s' )

    npt.assert_allclose( expected, actual, rtol=1e-4 )



#!/usr/bin/env python
'''Compilation of observed and derived astrophysical trends for CGMs.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import numpy as np
import os
import scipy.interpolate as interp

import galaxy_dive.config as gd_config

########################################################################

def cooling_time(
    r,
    z,
    sim_name,
    physics,
    data_dir = gd_config.DATA_DIR,
    filename = 'Rovertcool_res7100.npz',
):
    '''Extract the pre-calculated cooling time for a given snapshot
    from a look-up table.

    Args:
        r (float or array-like):
            Radii (in pkpc) at which to look up the cooling time.

        z (float or array-like):
            Redshifts at which to look up the cooling time.

        physics (str):
            What simulation variant to use (e.g. 'core' or 'metal_diffusion').

        data_dir (str):
            Directory containing the data.

        filename (str):
            Filename for lookup table.

    Returns:
        value or array-like (same shape as r and z):
            Interpolated values of t_cool from look-up table.
    '''

    assert sim_name[:3] == 'm12', \
        'Cooling time currently only calculated for m12 halos'

    # Load up the file
    filepath = os.path.join( data_dir, filename )
    data = np.load( filepath )

    # Translate the keys from bytes...
    keys = data['key'].astype( str )

    # Make sure the sim name exists in the sim
    assert sim_name[3] in keys[:,0]

    # Identify the relevant simulation snapshots
    matches_sim = keys[:,0] == sim_name[3]
    matches_physics = keys[:,1] == physics
    valid_snapshots = matches_sim & matches_physics

    # Get the redshifts stored in the file
    unsorted_redshifts = keys[:,2][valid_snapshots].astype( float )
    radii = data['Rs'][valid_snapshots]
    t_cools = data['tcools'][valid_snapshots]

    # Sort, tile, and mask
    sort_inds = unsorted_redshifts.argsort()
    redshift = unsorted_redshifts[sort_inds]
    redshift = np.tile( redshift, ( radii.shape[1], 1 ) ).transpose()
    radii = radii[sort_inds]
    t_cools = t_cools[sort_inds]

    # Interpolate
    return interp.griddata(
        ( redshift.flatten(), radii.flatten() ),
        t_cools.flatten(),
        ( z, r ),
    )
    



# -*- coding: utf-8 -*-
# -*- mode: python -*-
# Adapted from mpl_toolkits.axes_grid1
# LICENSE: Python Software Foundation (http://docs.python.org/license.html)

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.offsetbox import AnchoredOffsetbox
from matplotlib.ticker import MaxNLocator

########################################################################

class AnchoredScaleBar( AnchoredOffsetbox ):

    def __init__( self, transform, sizex=0, sizey=0, labelx=None, labely=None, loc=4,
                 pad=0.1, borderpad=0.1, sep=2, prop=None, xorder='text above', 
                 textprops=None, linewidth=3, bbox=[1, 1.115, 0, 0], bbox_transform=None, **kwargs ):
        """Draw a horizontal and/or vertical  bar with the size in data coordinate
        of the give axes. A label will be drawn underneath (center-aligned).
        
        This is from GitHubGist user dmeliza, with the code found at https://gist.github.com/dmeliza/3251476

        Args:
          transform : the coordinate frame (typically axes.transData)
          sizex,sizey : width of x,y bar, in data units. 0 to omit
          labelx,labely : labels for x,y bars; None to omit
          loc : position in containing axes
          pad, borderpad : padding, in fraction of the legend font size (or prop)
          sep : separation between labels and bars in points.
          **kwargs : additional arguments passed to base class constructor
          linewidth : Thickness of the scale bar
          bbox : Where the scale bar should be relative to the axis
        """
        from matplotlib.patches import Rectangle
        from matplotlib.offsetbox import AuxTransformBox, VPacker, HPacker, TextArea, DrawingArea
        bars = AuxTransformBox(transform)
        if sizex:
            bars.add_artist(Rectangle((0,0), sizex, 0, fc="none", linewidth=linewidth))

        if sizey:
            bars.add_artist(Rectangle((0,0), 0, sizey, fc="none"), linewidth=linewidth)

        if sizex and labelx:
            text_area = TextArea(labelx, textprops=textprops, minimumdescent=False)
            if xorder == 'text above':
              bar_child = [text_area, bars]
            elif xorder == 'text below':
              bar_child = [bars, text_area]
            bars = VPacker(children=bar_child, align="right", pad=0, sep=sep)
        if sizey and labely:
            text_area = TextArea(labely, textprops=textprops)
            bars = HPacker(children=[text_area, bars],
                            align="center", pad=0, sep=sep)

        AnchoredOffsetbox.__init__(self, loc, pad=pad, borderpad=borderpad,
                                   child=bars, prop=prop, frameon=False, 
                                   bbox_transform=bbox_transform, **kwargs)

        AnchoredOffsetbox.set_bbox_to_anchor(self, bbox, transform=bbox_transform)

def add_scalebar( ax, matchx=True, matchy=True, hidex=True, hidey=True, **kwargs ):
    """ Add scalebars to axes
    Adds a set of scale bars to *ax*, matching the size to the ticks of the plot
    and optionally hiding the x and y axes

    Args:
      ax : the axis to attach ticks to
      matchx,matchy : if True, set size of scale bars to spacing between ticks
                    if False, size should be set using sizex and sizey params
      hidex,hidey : if True, hide x-axis and y-axis of parent
      **kwargs : additional arguments passed to AnchoredScaleBars

    Returns:
      created scalebar object
    """
    def f(axis):
        l = axis.get_majorticklocs()
        return len(l)>1 and (l[1] - l[0])
    
    if matchx:
        kwargs['sizex'] = f(ax.xaxis)
        kwargs['labelx'] = str(kwargs['sizex'])
    if matchy:
        kwargs['sizey'] = f(ax.yaxis)
        kwargs['labely'] = str(kwargs['sizey'])
        
    sb = AnchoredScaleBar(ax.transData, **kwargs)
    ax.add_artist(sb)

    if hidex : ax.xaxis.set_visible(False)
    if hidey : ax.yaxis.set_visible(False)

    return sb




#!/usr/bin/env python
'''Non-physical constants for use in analyzing data.
For example, this contains the default fill values for invalid data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import numpy as np

########################################################################
########################################################################

INT_FILL_VALUE = -99999
FLOAT_FILL_VALUE = np.nan

PTYPES = {
  'gas' : 0,
  'DM' : 1,
  'lowresDM' : 2,
  'star' : 4,
}

STANDARD_PTYPES = [ 'gas', 'DM', 'lowresDM', 'star', ]



'''Testing for utilities.py
'''

from mock import patch
import numpy as np
import numpy.testing as npt
import os
import pdb
import shutil
import subprocess
import unittest

import galaxy_dive.utils.utilities as utilities

########################################################################
########################################################################

class TestDictFromDefaultsAndVariations( unittest.TestCase ):

    def test_default( self ):

        defaults = { 'best cat' : 'Melvulu', }
        variations = {
            'person a' : { 'other best cat' : 'Chellbrat', },
            'person b' : {
                'best cat' : 'A Normal Melville Cat',
                'other best cat' : 'Chellcat',
            },
        }

        actual = utilities.dict_from_defaults_and_variations( defaults, variations )
        expected = {
            'person a' : {
                'best cat' : 'Melvulu',
                'other best cat' : 'Chellbrat',
            },
            'person b' : {
                'best cat' : 'A Normal Melville Cat',
                'other best cat' : 'Chellcat',
            },
        }

        for key in expected.keys():
            assert expected[key] == actual[key]

########################################################################
########################################################################

class TestGetCodeVersion( unittest.TestCase ):

    def test_default( self ):

        result = utilities.get_code_version( self )

        assert result is not None

    ########################################################################

    def test_works_in_otherdir( self ):

        cwd = os.getcwd()

        # Change directories
        os.chdir( '../' )

        result = utilities.get_code_version( self )

        assert result is not None

        # Change back
        os.chdir( cwd )

    ########################################################################

    def test_works_for_modules( self ):

        actual = utilities.get_code_version( utilities, instance_type='module' )

        expected = subprocess.check_output( [ 'git', 'describe', '--always'] )

        assert actual == expected

    ########################################################################

    def test_doesnt_change_dir_upon_failing( self ):

        true_cwd = os.getcwd()

        # Change to a dir that should not be a repository
        os.chdir( os.path.expanduser( '~' ) )

        cwd = os.getcwd()

        result = utilities.get_code_version( self )

        now_cwd = os.getcwd()

        assert cwd == now_cwd

        # Change back
        os.chdir( true_cwd )

########################################################################
########################################################################

class TestGetInstanceSourceDir( unittest.TestCase ):

    def test_default( self ):

        actual = utilities.get_instance_source_dir( self )

        expected = os.path.join( os.getcwd(), 'tests/test_utils' )

        assert actual == expected

    ########################################################################

    def test_works_for_modules( self ):

        actual = utilities.get_instance_source_dir( utilities, 'module' )

        expected = os.path.join( os.getcwd(), 'utils' )

        assert actual == expected

########################################################################
########################################################################

class TestChunkList( unittest.TestCase ):

    def test_default( self ):

        n = 2
        l = range( 10 )
        expected = ( range(0,5), range(5,10) )
        actual = utilities.chunk_list( l, n )
        for expected_, actual_ in zip( expected, actual ):
            assert expected_ == actual_

        n = 3
        l = range( 10 )
        expected = ( range(0,4), range(4,7), range(7,10) )
        actual = utilities.chunk_list( l, n )
        for expected_, actual_ in zip( expected, actual ):
            assert expected_ == actual_

        n = 3
        l = range( 11 )
        expected = ( range(0,4), range(4,8), range(8,11) )
        actual = utilities.chunk_list( l, n )
        for expected_, actual_ in zip( expected, actual ):
            assert expected_ == actual_

########################################################################
########################################################################

class TestArraySetConversion( unittest.TestCase ):

    def test_arrays_to_set( self ):

        expected = set( [ ( 1, 1 ), ( 2, 2 ), ( 0, 0 ) ] )

        actual = utilities.arrays_to_set( np.arange( 3 ), np.arange( 3 ) )

        self.assertEqual( expected, actual )

    ########################################################################

    def test_arrays_to_set_long( self ):

        expected = set( [ ( 1, 1, 1 ), ( 2, 2, 2 ), ( 0, 0, 0 ) ] )

        actual = utilities.arrays_to_set( np.arange( 3 ), np.arange( 3 ), np.arange(3) )

        self.assertEqual( expected, actual )

    ########################################################################

    def test_set_to_arrays( self ):

        expected = np.array([
            [ 1, 4, 7, ],
            [ 2, 5, 8, ],
            [ 3, 6, 9, ],
        ])

        actual = utilities.set_to_arrays( set( [ ( 1, 2, 3, ), ( 4, 5, 6 ), ( 7, 8, 9 ), ] ) )

        npt.assert_allclose( expected[:,0], actual[:,-1] )

        # To allow for full comparison. We don't really care what order they're in, as long as they're matched.
        actual.sort()

        npt.assert_allclose( expected, actual )

########################################################################
########################################################################

class TestStoreParameters( unittest.TestCase ):

    def test_basic( self ):

        class Object( object ):

            @utilities.store_parameters
            def __init__( self, a, b, c=3 ):
                pass

        o = Object( 1, 2, )

        self.assertEqual( 1, o.a )
        self.assertEqual( 2, o.b )
        self.assertEqual( 3, o.c )
        self.assertEqual( [ 'a', 'b', 'c' ], o.stored_parameters )

    ########################################################################

    def test_kwargs_input( self ):

        class Object( object ):

            @utilities.store_parameters
            def __init__( self, a, b, c=3, **kwargs ):
                pass

        o = Object( **{ 'a' : 1, 'b' : 2, 'c' : 4, 'd' : 5 } )

        self.assertEqual( 1, o.a )
        self.assertEqual( 2, o.b )
        self.assertEqual( 4, o.c )
        self.assertEqual( {'d' : 5}, o.kwargs )
        self.assertEqual(
            sorted( [ 'a', 'b', 'c', 'kwargs', ] ),
            sorted( o.stored_parameters )
        )

    ########################################################################

    def test_args_and_kwargs_input( self ):

        class Object( object ):

            @utilities.store_parameters
            def __init__( self, *args, **kwargs ):
                pass

        args = ( 'dog', )
        kwargs = { 'a' : 1, 'b' : 2, 'c' : 4, 'd' : 5 }
        o = Object( *args, **kwargs )

        self.assertEqual( ('dog',), o.args )
        self.assertEqual( { 'a' : 1, 'b' : 2, 'c' : 4, 'd' : 5 }, o.kwargs )
        self.assertEqual( [ 'args', 'kwargs', ], o.stored_parameters )

########################################################################
########################################################################

class TestGenerateDerivedDataDoc( unittest.TestCase ):

    def setUp( self ):

        self.output_filepath = './tests/data/derived_data.rst'

        if os.path.isfile( self.output_filepath ):
            os.remove( self.output_filepath )

    ########################################################################

    def test_gen_derived_data_doc( self ):

        object_import_path = 'galaxy_dive.tests.test_utils.test_utilities.GenDerivedDataObject'

        # Call the function
        utilities.gen_derived_data_doc(
            object_import_path,
            self.output_filepath,
        )
                
        # Build Expected
        # Get the header
        with open ( './utils/derived_data_header.rst', 'r') as myfile:
            expected_header = myfile.readlines()
        # Get the generated data
        expected_generated = [
            '\n',
            '.. autoclass:: {}\n'.format( object_import_path ),
            '    :show-inheritance:\n',
            '\n',
            '    .. automethod:: calc_A\n',
            '    .. automethod:: calc_B\n',
        ]
        expected = expected_header + expected_generated

        with open ( self.output_filepath, 'r') as myfile:
            actual = myfile.readlines()

        self.assertEqual( expected, actual )

# Test Object for TestGenerateDerivedDataDoc
class GenDerivedDataObject( object ):

    def method_a( self, *args, **kwargs ):
        '''Method a.'''
        pass

    def calc_A( self ):
        '''Calculate A''' 
        pass

    def calc_B( self ):
        '''Calculate A''' 
        pass



#!/usr/bin/env python
'''Code for managing data troves.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import itertools
import os

import galaxy_dive.utils.utilities as utilities

########################################################################
########################################################################


class TroveManager( object ):
    '''Class for managing troves of data.'''

    @utilities.store_parameters
    def __init__( self, file_format, *args ):
        '''Constructor.

        Args:
            file_format (str) :
                Format for data files.

            *args :
                Arguments to pass to self.get_file() to get different data files.

        Returns:
            TroveManager object.
        '''

        pass

    ########################################################################

    def get_file( self, *args ):
        '''Default method for getting the data filename.
        
        Args:
            *args :
                Arguments provided. Assumes args[0] is the data dir.

        Returns:
            Filename for a given combination of args.
        '''

        filename = self.file_format.format( *args[1:] )

        return os.path.join( args[0], filename )

    ########################################################################

    @property
    def combinations( self ):
        '''Returns:
            All combinations of arguments.
        '''

        if not hasattr( self, '_combinations' ):
            self._combinations = list( itertools.product( *self.args ) )

        return self._combinations

    ########################################################################

    @property
    def data_files( self ):
        '''Returns:
            All data files that should be part of the trove.
        '''

        if not hasattr( self, '_data_files' ):
            self._data_files = [
                self.get_file( *args ) for args in self.combinations
             ]

        return self._data_files

    ########################################################################

    def get_incomplete_combinations( self ):
        '''Returns:
            Combinations in the trove that have not yet been done.
        '''

        incomplete_combinations = []
        for i, data_file in enumerate( self.data_files ):

            if not os.path.isfile( data_file ):
                incomplete_combinations.append( self.combinations[i] )

        return incomplete_combinations

    ########################################################################

    def get_incomplete_data_files( self ):
        '''Returns:
            Data files in the trove that have not yet been done.
        '''

        return [
            self.get_file( *args ) for args \
                in self.get_incomplete_combinations()
        ]

    ########################################################################

    def get_next_args_to_use( self, when_done='return_last' ):
        '''Is this necessary? No. This function is really a wrapper that in
        essence provides documentation.


        Args:
            when_done (str) :
                What to do when there are no incomplete combinations? Defaults
                to returning the last of self.combinations.

        Returns:
            Next set of arguments to use.
        '''

        incomplete_combinations = self.get_incomplete_combinations()

        if len( incomplete_combinations ) == 0:
            if when_done == 'return_last':
                return self.combinations[-1]
            elif when_done == 'return_0':
                return 0

        return self.get_incomplete_combinations()[0]



'''Testing for mp_utils.py
'''

from mock import patch
import multiprocessing as mp
import numpy as np
import numpy.testing as npt
import os
import pdb
import subprocess
import unittest

import galaxy_dive.utils.mp_utils as mp_utils

########################################################################
########################################################################

class TestApplyAmongProcessors( unittest.TestCase ):

  def test_default( self ):

    d = {
      'a' : mp.Array( 'f', np.zeros( 10 ) ),
      'b' : mp.Array( 'f', np.zeros( 10 ) ),
    }
    all_args = [ ( i, d ) for i in range( 10 ) ]

    def f( i, d ):

      num = float( i + 1 )

      d['a'][i] = num**2.
      d['b'][i] = num*2.

    mp_utils.apply_among_processors( f, all_args, 2 )

    expected_d = {
      'a' : np.arange( 1, 11, dtype=float)**2.,
      'b' : np.arange( 1, 11, dtype=float)*2.,
    }

    for key in expected_d.keys():
      expected = expected_d[key]
      actual = d[key]

      npt.assert_allclose( expected, actual )
    
########################################################################
########################################################################

class TestMPQueueToList( unittest.TestCase ):

  def test_default( self ):

    q = mp.Queue()
    [ q.put( i ) for i in range( 4 ) ]
    
    actual = mp_utils.mp_queue_to_list( q, 2 )
    expected = range( 4 )

    npt.assert_allclose( sorted( expected ), sorted( actual ) )

  ########################################################################

  def test_long_odd_list( self ):

    q = mp.Queue()
    [ q.put( i ) for i in range( 7 ) ]
    
    actual = mp_utils.mp_queue_to_list( q, 2 )
    expected = range( 7 )

    npt.assert_allclose( sorted( expected ), sorted( actual ) )

  ########################################################################

  def test_more_processors( self ):

    q = mp.Queue()
    [ q.put( i ) for i in range( 2 ) ]
    
    actual = mp_utils.mp_queue_to_list( q, 4 )
    expected = range( 2 )

    npt.assert_allclose( sorted( expected ), sorted( actual ) )

  ########################################################################

  def test_single_object( self ):

    q = mp.Queue()
    [ q.put( i ) for i in range( 1 ) ]
    
    actual = mp_utils.mp_queue_to_list( q, 2 )
    expected = range( 1 )

    npt.assert_allclose( expected, actual )

  ########################################################################

  def test_queue_of_sets( self ):

    q = mp.Queue()

    ls = [ range( i, i + 4 ) for i in range( 3 ) ]
    sets = [ set( l ) for l in ls ]
    [ q.put( s ) for s in sets ]
    
    actual = mp_utils.mp_queue_to_list( q, 2 )
    expected = sets

    actual_sorted = sorted( [ list( s ) for s in actual ] )
    expected_sorted = sorted( [ list( s ) for s in expected ] )

    npt.assert_allclose( expected_sorted, actual_sorted )



#!/usr/bin/env python
'''Utilities for multiprocessing.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import multiprocessing as mp
import os
import pdb
import sys
from types import MethodType

# Python 2/3 compatible copyreg
try:
    import copy_reg
except:
    import copyreg as copy_reg

import galaxy_dive.utils.utilities as utilities

########################################################################

def apply_among_processors( f, all_args, n_processors=mp.cpu_count() ):
  '''Takes a list of arguments and breaks it up and splits those chunks among the processors.
  Note: This currently does not return anything, so it doesn't work for functions where you want f to return something.
  However! It does work for shared memory objects, unlike Pool or parmap!

  Args:
    f (function) : The function to apply the args to.
    all_args (list) : Args to apply. Format, [ (args1), (args2), ... ]
    n_processors (int, optional) : Number of processors to use.
  '''

  def wrapped_f( args_chunk ):
    for args in args_chunk:
      f(*args)

  chunked_args = utilities.chunk_list( all_args, n_processors )

  ps = [ mp.Process( target=wrapped_f, args=(args_chunk,) ) for args_chunk in chunked_args ]

  [ p.start() for p in ps ]
  [ p.join() for p in ps ]

########################################################################

def mp_queue_to_list( queue, n_processors=mp.cpu_count() ):
  '''Convert a multiprocessing.Queue object to a list, using multiple processors to parse it.
  The list is unordered. It may also not work if the queue contains lists.

  Args:
    queue (mp.Queue) : The queue to turn into a list.
    n_processors (int) : Number of processors to use.
  '''

  def process_queue( q, l ):

    while True:

      l.acquire()
      if q.qsize() > 1:
        popped1 =  q.get()
        popped2 =  q.get()
        l.release()
      else:
        l.release()
        break

      if not isinstance( popped1, list ):
        popped1 = [ popped1, ]
      if not isinstance( popped2, list ):
        popped2 = [ popped2, ]
      
      q.put( popped1 + popped2 )

  lock = mp.Lock()

  proc = [ mp.Process( target=process_queue, args=(queue,lock) ) for _ in range( n_processors ) ]
  for p in proc:
      p.daemon = True
      p.start()
  [ p.join() for p in proc ]

  return queue.get()

########################################################################
'''The following is a version of Pool, written with classes in mind. It does not handle shared memory objects well.
https://stackoverflow.com/a/16071616
'''

def fun( f, q_in, q_out ):
  while True:
    i, x = q_in.get()
    if i is None:
      print( "PID {} finishing, PPID {}.".format( os.getpid(), os.getppid() ) )
      break
    q_out.put( (i, f( x )) )

def set_fun( f, q_in, q_out ):
  res_proc = set()
  while True:
    i, x = q_in.get()
    if i is None:
      print( "PID {} finishing, PPID {}.".format( os.getpid(), os.getppid() ) )
      q_out.put( res_proc )
      break
    res_proc = res_proc | f( x )

def parmap( f, X, n_processors=mp.cpu_count(), return_values=True, set_case=False, use_mp_queue_to_list=False ):
    '''Parallel map, viable with classes.

    Args:
      f (function) : Function to map to.
      X (list) : List of arguments to provide f
      n_processors (int) : Number of processors to use.
      return_values (bool) : If False, don't bother getting the results from the functions.
      set_case (bool) : If this option is True, it assumes that f returns a set, and that results should be the
        union of all those sets.
      use_mp_queue_to_list (bool) : Experimental. If True, try to use mp_queue_to_list to convert the list.
        Only works if set_case, currently.

    Returns:
      results (list or set) : The results.
    '''

    m = mp.Manager()

    q_in = m.Queue(1)
    q_out = m.Queue()

    if set_case:
      target_fun = set_fun
    else:
      target_fun = fun

    proc = [ mp.Process( target=target_fun, args=(f, q_in, q_out) )
            for _ in range( n_processors ) ]
    for p in proc:
        p.daemon = True
        p.start()

    sent = [ q_in.put( (i, x) ) for i, x in enumerate( X ) ]
    [ q_in.put( (None, None) ) for _ in range( n_processors ) ]

    print( "Getting results from queue. This could take a while..." )

    # Store the results
    if return_values:
      if set_case:

        if use_mp_queue_to_list:
          res = mp_queue_to_list( q_out, n_processors )
        else:
          res = [ q_out.get() for _ in range( n_processors ) ]

        [ p.join() for p in proc ]

        return res

      else:

        res = [ q_out.get() for _ in range( len( sent ) ) ]

        [ p.join() for p in proc ]

        return [ x for i, x in sorted( res ) ]

    else:
      [ p.join() for p in proc ]

      return

########################################################################
'''This section contains efforts to make classes pickleable, allowing multiprocessing.Pool to be used.'''

def _pickle_method(method):
  '''The majority of this was taken from the following StackOverflow answer:
  http://stackoverflow.com/questions/1816958/cant-pickle-type-instancemethod-when-using-pythons-multiprocessing-pool-ma/7309686#7309686
  '''

  func_name = method.im_func.__name__
  obj = method.im_self
  cls = method.im_class
  return _unpickle_method, (func_name, obj, cls)

def _unpickle_method(func_name, obj, cls):
  '''The majority of this was taken from the following StackOverflow answer:
  http://stackoverflow.com/questions/1816958/cant-pickle-type-instancemethod-when-using-pythons-multiprocessing-pool-ma/7309686#7309686
  '''

  for cls in cls.mro():
    try:
      func = cls.__dict__[func_name]
    except KeyError:
      pass
    else:
      break
  return func.__get__(obj, cls)

def make_classes_picklable():
  '''The majority of this was taken from the following StackOverflow answer:
  http://stackoverflow.com/questions/1816958/cant-pickle-type-instancemethod-when-using-pythons-multiprocessing-pool-ma/7309686#7309686
  '''
  copy_reg.pickle(MethodType, _pickle_method, _unpickle_method)

########################################################################

class ForkedPdb(pdb.Pdb):
    """A Pdb subclass that may be used
    from a forked multiprocessing child
    From https://stackoverflow.com/a/23654936
    """
    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            sys.stdin = open('/dev/stdin')
            pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin



#!/usr/bin/env python
'''Test suite config file.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import pytest

########################################################################

def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true",
        help="run slow tests")
    parser.addoption("--runstampede", action="store_true",
        help="run tests unique to stampede")



#!/usr/bin/env python
'''Tools for modifying halo data output files.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import copy
import glob
import numpy as np
import os
import pandas as pd

import galaxy_dive.galaxy_linker.linker as galaxy_linker
import galaxy_dive.analyze_data.halo_data as halo_data
import galaxy_dive.read_data.metafile as read_metafile
import galaxy_dive.utils.astro as astro_utils
import galaxy_dive.utils.constants as constants
import galaxy_dive.utils.data_constants as data_constants
import galaxy_dive.utils.data_operations as data_operations
import galaxy_dive.utils.utilities as utilities

import galaxy_dive.analyze_data.ahf as ahf
import galaxy_dive.analyze_data.particle_data as particle_data

########################################################################
########################################################################

class HaloUpdater( halo_data.HaloData ):
    '''Class for updating Halo data (smoothing, adding in additional columns, etc)'''

    def __init__( self, *args, **kwargs ):

        self.key_parser = ahf.HaloKeyParser()

        super( HaloUpdater, self ).__init__( *args, **kwargs )

    ########################################################################
    # Get Data Values
    ########################################################################

    def get_accurate_redshift( self, metafile_dir ):
        '''Get a better values of the redshift than what's stored in the Halo filename, by loading them from an external file.

        Args:
            metafile_dir (str): The directory the snapshot_times are stored in.

        Modifies:
            self.mtree_halos (dict of pd.DataFrames): Updates the redshift column
        '''

        # Get the redshift data out
        metafile_reader = read_metafile.MetafileReader( metafile_dir )
        metafile_reader.get_snapshot_times()

        # Replace the old data
        for halo_id in self.mtree_halos.keys():

            mtree_halo = self.mtree_halos[ halo_id ]

            # Read and replace the new redshift
            new_redshift = metafile_reader.snapshot_times['redshift'][ mtree_halo.index ]
            mtree_halo['redshift'] = new_redshift

    ########################################################################

    def get_analytic_concentration( self, metafile_dir, type_of_halo_id='merger_tree' ):
        '''Get analytic values for the halo concentration, using colossus, Benedikt Diemer's cosmology code
        ( https://bitbucket.org/bdiemer/colossus ; http://www.benediktdiemer.com/code/colossus/ ).

        Assumptions:
            - We're using the default formula of Diemer&Kravtstov15
            - We're using the Bryan&Norman1998 version of the virial radius.

        Args:
            metafile_dir (str): The directory the snapshot_times are stored in.
            type_of_halo_id (str): 'merger_tree' if the halo id is a merger tree halo id.
                                                          'halos' if the halo id is a *.AHF_halos halo id.

        Returns
            c_vir (np.array of floats): The concentration, defined as R_vir/r_scale.
        '''

        # Include imports here, because this function may not in general work if colossus is not available,
        # and the rest of the module should still be made useable
        # There may be some warnings here about the version of scipy colossus uses, as opposed to the version galaxy_dive uses
        import colossus.cosmology.cosmology as co_cosmology
        import colossus.halo.concentration as co_concentration

        # Get simulation parameters, for use in creating a cosmology
        metafile_reader = read_metafile.MetafileReader( metafile_dir )
        metafile_reader.get_used_parameters()

        # Setup the cosmology used by the simulations
        sim_cosmo = {
            'flat': True,
            'H0' : float( metafile_reader.used_parameters['HubbleParam'] )*100.,
            'Om0' : float( metafile_reader.used_parameters['Omega0'] ),
            'Ob0' : float( metafile_reader.used_parameters['OmegaBaryon'] ),
            'sigma8' : co_cosmology.cosmologies['WMAP9']['sigma8'], # Use WMAP9 for values we don't store in our simulations explicitly.
            'ns' : co_cosmology.cosmologies['WMAP9']['ns'], # Use WMAP9 for values we don't store in our simulations explicitly.
        }
        cosmo = co_cosmology.setCosmology( 'sim_cosmo', sim_cosmo )

        if type_of_halo_id == 'merger_tree':

            # Loop over all mt halos
            for halo_id in self.mtree_halos.keys():

                # Load the data
                mtree_halo = self.mtree_halos[ halo_id ]

                # Get the concentration out
                c_vir = []
                for m_vir, z in zip( mtree_halo['Mvir'], mtree_halo['redshift'] ):
                    c = co_concentration.concentration( m_vir, 'vir', z, model='diemer15', statistic='median')
                    c_vir.append( c )

                # Turn the concentration into an array
                c_vir = np.array( c_vir )

                # Save the concentration
                mtree_halo['cAnalytic'] = c_vir

        elif type_of_halo_id == 'halos':

            # Get the redshift for the halo file.
            metafile_reader.get_snapshot_times()
            redshift = metafile_reader.snapshot_times['redshift'][self.halos_snum]

            # Get the concentration
            c = co_concentration.concentration( self.halos['Mvir'], 'vir', redshift, model='diemer15', statistic='median')

            return c

    ########################################################################

    def get_mass_radii(
                self,
                mass_fractions,
                simulation_data_dir,
                galaxy_cut,
                length_scale,
        ):
        '''Get radii that enclose a fraction (mass_fractions[i]) of a halo's stellar mass.

        Args:
            mass_fractions (list of floats) :
                Relevant mass fractions.

            simulation_data_dir (str) :
                Directory containing the raw particle data.

            galaxy_cut (float) :
                galaxy_cut*length_scale defines the radius around the center of the halo to look for stars.

            length_scale (str) :
                galaxy_cut*length_scale defines the radius around the center of the halo to look for stars.

        Returns:
            mass_radii (list of np.ndarrays) :
                If M_sum_j = all mass inside galaxy_cut*length_scale for halo j, then mass_radii[i][j] is the radius that
                contains a fraction mass_fractions[i] of M_sum_j.
        '''
                
        # Load the simulation data
        s_data = particle_data.ParticleData(
            simulation_data_dir,
            self.halos_snum,
            ptype = data_constants.PTYPES['star'],
        )

        try:
            particle_positions = s_data.data['P'].transpose()
        # Case where there are no star particles at this redshift.
        except KeyError:
            return [ np.array( [ np.nan, ]*self.halos.index.size ), ]*len( mass_fractions )

        # Find the mass radii
        galaxy_linker_kwargs = {
            'particle_positions' : particle_positions,
            'particle_masses' : s_data.data['M'],
            'snum' : self.halos_snum,
            'redshift' : s_data.redshift,
            'hubble' : s_data.data_attrs['hubble'],
            'galaxy_cut' : galaxy_cut,
            'length_scale' : length_scale,
            'halo_data' : self,
        }
        gal_linker = galaxy_linker.GalaxyLinker( **galaxy_linker_kwargs )
        
        mass_radii = [ gal_linker.get_mass_radius( mass_fraction ) for mass_fraction in mass_fractions ]
            
        return mass_radii

    ########################################################################

    def get_enclosed_mass( self,
        simulation_data_dir,
        ptype,
        galaxy_cut,
        length_scale,
        ):
        '''Get the mass inside galaxy_cut*length_scale for each Halo halo.

        Args:
            simulation_data_dir (str) :
                Directory containing the raw particle data.

            ptype (str) :
                What particle type to get the mass for.

            galaxy_cut (float) :
                galaxy_cut*length_scale defines the radius around the center of the halo within which to get the mass.

            length_scale (str) :
                galaxy_cut*length_scale defines the radius around the center of the halo within which to get the mass.

        Returns:
            mass_inside_all_halos (np.ndarray) :
                mass_inside_all_halos[i] is the mass of particle type ptype inside galaxy_cut*length scale around a galaxy.
        '''

        # Load the simulation data
        s_data = particle_data.ParticleData(
            simulation_data_dir,
            self.halos_snum,
            data_constants.PTYPES[ptype],
        )

        try:
            particle_positions = s_data.data['P'].transpose()
        # Case where there are no star particles at this redshift.
        except KeyError:
            return np.array( [ 0., ]*self.halos.index.size )

        # Find the mass radii
        galaxy_linker_kwargs = {
            'particle_positions' : particle_positions,
            'particle_masses' : s_data.data['M']*constants.UNITMASS_IN_MSUN,
            'snum' : self.halos_snum,
            'redshift' : s_data.redshift,
            'hubble' : s_data.data_attrs['hubble'],
            'galaxy_cut' : galaxy_cut,
            'length_scale' : length_scale,
            'halo_data' : self,
        }
        gal_linker = galaxy_linker.GalaxyLinker( **galaxy_linker_kwargs )

        mass_inside_all_halos = gal_linker.mass_inside_all_halos
        
        # Make sure to put hubble constant back in so we have consistent units.
        mass_inside_all_halos *= s_data.data_attrs['hubble']

        return mass_inside_all_halos

    ########################################################################

    def get_average_quantity_inside_galaxy( self,
        data_key,
        simulation_data_dir,
        ptype,
        galaxy_cut,
        length_scale,
        weight_data_key = 'M',
        fill_value = np.nan,
        ):
        '''Get the mass inside galaxy_cut*length_scale for each Halo halo.

        Args:
            data_key (str) :
                Data key for the quantity to get the average of.

            simulation_data_dir (str) :
                Directory containing the raw particle data.

            ptype (str) :
                What particle type to get the mass for.

            galaxy_cut (float) :
                galaxy_cut*length_scale defines the radius around the center of the halo within which to get the mass.

            length_scale (str) :
                galaxy_cut*length_scale defines the radius around the center of the halo within which to get the mass.

            weight_data_key (str) :
                Data key for the weight to use when averaging.

            fill_value (float) :
                What value to use when the average quantity inside the galaxy is not resolved.

        Returns:
            average_quantity_inside_galaxy (np.ndarray) :
                average_quantity_inside_galaxy[i] is the average value of the requested quantity for particle type ptype
                inside galaxy_cut*length scale around a galaxy.
        '''

        # Load the simulation data
        s_data = particle_data.ParticleData(
            simulation_data_dir,
            self.halos_snum,
            data_constants.PTYPES[ptype],

            # The following values need to be set, because they come into play when a galaxy is centered on halo finder
            # data. That's obviously not the case here...
            centered = True,
            vel_centered = True,
            hubble_corrected = True,
        )

        try:
            particle_positions = s_data.data['P'].transpose()
        # Case where there are no particles of the given ptype at this redshift.
        except KeyError:
            return np.array( [ fill_value, ]*self.halos.index.size )

        # Find the mass radii
        galaxy_linker_kwargs = {
            'particle_positions' : particle_positions,
            'snum' : self.halos_snum,
            'redshift' : s_data.redshift,
            'hubble' : s_data.data_attrs['hubble'],
            'galaxy_cut' : galaxy_cut,
            'length_scale' : length_scale,
            'halo_data' : self,
        }
        gal_linker = galaxy_linker.GalaxyLinker( low_memory_mode=False, **galaxy_linker_kwargs )

        average_quantity_inside_galaxy = gal_linker.weighted_summed_quantity_inside_galaxy(
            s_data.get_data( data_key ),
            s_data.get_data( weight_data_key ),
            fill_value,
        )

        return average_quantity_inside_galaxy

    ########################################################################
    
    def get_circular_velocity( self,
        galaxy_cut,
        length_scale,
        metafile_dir,
        ptypes = data_constants.STANDARD_PTYPES,
        ):
        '''Get the circular velocity at galaxy_cut*length_scale.

        Args:
            galaxy_cut (float) :
                galaxy_cut*length_scale defines the radius around the center of the halo within which to get the mass.

            length_scale (str) :
                galaxy_cut*length_scale defines the radius around the center of the halo within which to get the mass.

            metafile_dir (str) :
                Directory containing metafile data, for getting out the redshift given a snapshot.

            ptypes (list of strs) :
                Particle types to count the mass inside the halo of.

        Returns:
            v_circ (np.ndarray) 
                Circular velocity at galaxy_cut*length_scale using mass from the given ptypes.
        '''

        # Get the redshift, for converting the radius to pkpc/h.
        metafile_reader = read_metafile.MetafileReader( metafile_dir )
        metafile_reader.get_snapshot_times()
        redshift = metafile_reader.snapshot_times['redshift'][self.halos_snum]

        # Get the radius in pkpc/h
        try:
            radius = galaxy_cut*self.halos[length_scale]
        except KeyError:
            radius = galaxy_cut*self.halos_add[length_scale]
        radius /= ( 1. + redshift )

        # Get the mass in Msun/h
        masses = []
        for ptype in ptypes:
            mass_key = self.key_parser.get_enclosed_mass_key( ptype, galaxy_cut, length_scale )
            try:
                ptype_mass = self.halos_add[mass_key]
            except:
                ptype_mass = self.halos[mass_key]
            masses.append( ptype_mass )
        mass = np.array( masses ).sum( axis=0 )

        # Now get the circular velocity out
        # (note that we don't need to bother with converting out the 1/h's, because in this particular case they'll cancel)
        v_circ = astro_utils.circular_velocity( radius, mass )

        return v_circ

    ########################################################################
    # Alter Data
    ########################################################################

    def smooth_mtree_halos(
            self,
            metafile_dir,
            keys_to_smooth = [],
            smooth_kwargs = { 'window_len' : 20, 'window' : 'flat' },
        ):
        '''Make Rvir and Mvir monotonically increasing, to help mitigate artifacts in the Halo-calculated merger tree.
        NOTE: This smooths in *physical* coordinates, so it may not be exactly smooth in comoving coordinates.

        Args:
            metafile_dir (str) :
                The directory the snapshot_times are stored in.

                keys_to_smooth (list of strs) :
                        If given, also smooth the data given by these keys.
                        This smoothing isn't done to assume a monotonic increase, but is
                        a convolve with a moving filter through data_operations.smooth()

                smooth_kwargs (dict) :
                        Specific arguments that determine exactly how the smoothing is
                        done, when also smoothing for specific keys.

        Modifies:
            self.mtree_halos (dict of pd.DataFrames) :
                Changes self.mtree_halos[halo_id]['Rvir'] and self.mtree_halos[halo_id]['Mvir'] to be monotonically increasing.
        '''

        # We need to get an accurate redshift in order to smooth properly
        self.get_accurate_redshift( metafile_dir )

        for halo_id in self.mtree_halos.keys():

            # Load the data
            mtree_halo = self.mtree_halos[ halo_id ]

            # Convert into physical coords for smoothing (we'll still leave the 1/h in place)
            r_vir_phys = mtree_halo['Rvir']/( 1. + mtree_halo['redshift'] )

            # Smooth r_vir
            r_vir_phys_smooth = np.maximum.accumulate( r_vir_phys[::-1] )[::-1]

            # Convert back into comoving and save
            mtree_halo['Rvir'] = r_vir_phys_smooth*( 1. + mtree_halo['redshift'] )

            # Smooth Mvir
            mtree_halo['Mvir'] = np.maximum.accumulate( mtree_halo['Mvir'][::-1] )[::-1]

            for smooth_key in keys_to_smooth:

                original_data = copy.copy( mtree_halo[smooth_key].values )

                smoothed_data = data_operations.smooth(
                        original_data,
                        **smooth_kwargs
                )

                # Replace NaN values with original values, where possible
                smoothed_nan = np.isnan( smoothed_data )
                smoothed_data[smoothed_nan] = original_data[smoothed_nan]

                smooth_save_key = 's' + smooth_key

                mtree_halo[smooth_save_key] = smoothed_data

    ########################################################################

    def include_halos_to_mtree_halos( self ):
        '''While most of the halofile data are contained in *.AHF_halos files, some quantities are stored in
        *.AHF_halos files. These are usually computed manually, external to what's inherent in AHF. This routine adds
        on the the information from these files to the loaded merger tree data (which don't usually include them, because
        they're not inherent to AHF.

        Modifies:
            self.mtree_halos (dict of pd.DataFrames) :
                Adds additional columns contained in *.AHF_halos_add files.
        '''

        for mtree_id, mtree_halo in self.mtree_halos.items():

            print( "Looking at merger tree ID {}".format( mtree_id ) )

            halo_ids = mtree_halo['ID'].values
            snums = mtree_halo.index

            ahf_frames = []
            for snum, halo_id in zip( snums, halo_ids, ):

                print( "Getting data for snapshot {}".format( snum ) )

                self.get_halos( snum )

                # Get the columns we want to add on.
                halofile_columns = set( self.halos.columns )
                mtree_columns = set( mtree_halo.columns )
                columns_to_add = list( halofile_columns - mtree_columns )
                columns_to_add.sort()

                # Now get the values to add
                if self.halos.index.size != 0:
                    full_ahf_row = self.halos.loc[halo_id:halo_id] 
                    ahf_row = full_ahf_row[columns_to_add]

                # Check for edge case, where there isn't an AHF row with specified halo number or there are no more halos
                if ( self.halos.index.size == 0 ) or ( ahf_row.size == 0 ):
                    
                    # Borrow a previous row for formatting
                    ahf_row = copy.copy( ahf_frames[-1] )

                    # Drop the previous row
                    ahf_row = ahf_row.drop( ahf_row.index[0] )

                    # Turn all the values to np.nan, so we know that they're invalid, but the format still works.
                    [ ahf_row.set_value( halo_id, column_name, np.nan ) for column_name in columns_to_add ]

                ahf_frames.append( ahf_row )

            custom_mtree_halo = pd.concat( ahf_frames )

            # Add in the snapshots, and use them as the index
            custom_mtree_halo['snum'] = snums
            custom_mtree_halo = custom_mtree_halo.set_index( 'snum', )

            # Now merge onto the mtree_halo DataFram
            self.mtree_halos[mtree_id] = pd.concat( [ mtree_halo, custom_mtree_halo, ], axis=1 )

    ########################################################################
    # Save Data
    ########################################################################

    def save_mtree_halos( self, tag ):
        '''Save loaded mergertree halo files in a csv file.

        Args:
            tag (str) : If the previous file was for example '/path/to/file/halo_00000.dat',
                                    the new file will be '/path/to/file/halo_00000_{}.dat'.format( tag )
        '''

        for halo_id in self.mtree_halos.keys():

            # Load the data
            mtree_halo = self.mtree_halos[ halo_id ]
            halo_filepath = self.mtree_halo_filepaths[ halo_id ]

            # Create the new filename
            filepath_base, file_ext = os.path.splitext( halo_filepath )
            save_filepath = '{}_{}{}'.format( filepath_base, tag, file_ext )

            mtree_halo.to_csv( save_filepath, sep='\t' )

    ########################################################################

    def save_smooth_mtree_halos(
            self,
            metafile_dir,
            index = None,
            include_halos_add = True,
            include_concentration = False,
            smooth_keys = [ 'Rstar0.5', ],
            **get_mtree_halo_kwargs
        ):
        '''Load halo files, smooth them, and save as a new file e.g., halo_00000_smooth.dat

        Args:
            metafile_dir (str) :
                The directory the metafiles (snapshot_times and used_parameters) are stored in.

            index (str or int) :
                What type of index to use. Defaults to None, which raises an exception. You *must* choose an
                index, to avoid easy mistakes. See get_mtree_halos() for a full description.

            include_concentration (bool):
                Whether or not to add an additional column that gives an analytic value for the
                halo concentration.
        '''
        
        # Load the data
        self.get_mtree_halos( index=index, **get_mtree_halo_kwargs )

        # Include data stored in *AHF_halos_add files.
        if include_halos_add:
            self.include_halos_to_mtree_halos()

        # Smooth the halos
        self.smooth_mtree_halos( metafile_dir, smooth_keys, )

        # Include the concentration, if chosen.
        if include_concentration:
            self.get_analytic_concentration( metafile_dir )

        # Save the halos
        self.save_mtree_halos( 'smooth' )

    ########################################################################

    def save_custom_mtree_halos( self, snums, halo_ids, metafile_dir, ):
        '''Save a custom merger tree.

        Args:
            snums (array-like or int) :
                What snapshots to generate the custom merger tree for.
                If a single integer, then snums will start at that integer and count backwards by single snapshots for the
                length of halo_ids

            halo_ids (array-like) :
                halo_ids[i] is the AHF_halos halo ID for the merger tree halo at snums[i].

            metafile_dir (str) :
                Directory for the metafile (used to get simulation redshift).

        Modifies:
            self.data_dir/halo_00000_custom.dat (text file) : Saves the custom merger tree at this location.
        '''

        if isinstance( snums, int ):
            snums = np.arange( snums, snums - len( halo_ids ), -1 )

        # Concatenate the data
        ahf_frames = []
        for snum, halo_id in zip( snums, halo_ids, ):

            print( "Getting data for snapshot {}".format( snum ) )

            self.get_halos( snum )

            ahf_frames.append( self.halos.loc[halo_id:halo_id] )

        custom_mtree_halo = pd.concat( ahf_frames )

        # Make sure to store the IDs too
        custom_mtree_halo['ID'] = halo_ids

        # Add in the snapshots, and use them as the index
        custom_mtree_halo['snum'] = snums
        custom_mtree_halo = custom_mtree_halo.set_index( 'snum', )

        # Get and save the redshift
        metafile_reader = read_metafile.MetafileReader( metafile_dir )
        metafile_reader.get_snapshot_times()
        custom_mtree_halo['redshift'] = metafile_reader.snapshot_times['redshift'][snums]

        # Save the data
        save_filepath = os.path.join( self.data_dir, 'halo_00000_custom.dat' )
        custom_mtree_halo.to_csv( save_filepath, sep='\t' )

    ########################################################################

    def save_halos_add( self,
        snum,
        include_analytic_concentration = True,
        include_mass_radii = True,
        include_enclosed_mass = True,
        include_average_quantity_inside_galaxy = False,
        include_v_circ = True,
        metafile_dir = None,
        simulation_data_dir = None,
        mass_radii_kwargs = {
            'mass_fractions' : [ 0.5, 0.75, 0.9, ],
            'galaxy_cut' : 0.15,
            'length_scale' : 'Rvir',
        },
        enclosed_mass_ptypes = data_constants.STANDARD_PTYPES,
        enclosed_mass_kwargs = {
            'galaxy_cut' : 5.0,
            'length_scale' : 'Rstar0.5',
        },
        average_quantity_data_keys = [ 'Vx', 'Vy', 'Vz', ],
        average_quantity_inside_galaxy_kwargs = {
            'ptype' : 'star',
            'galaxy_cut' : 5.0,
            'length_scale' : 'Rstar0.5',
        },
        v_circ_kwargs = {
            'galaxy_cut' : 5.0,
            'length_scale' : 'Rstar0.5',
        },
        verbose = False,
        ):
        '''Save additional columns that would be part of *.AHF_halos files, if that didn't break AHF.

        Args:
            snum (int) :
                Snapshot number to load.

            include_analytic_concentration (bool) :
                Include analytic concentration as one of the columns?

            include_mass_radii (bool) :
                Include radius that include some fraction of a particle's mass as one of the columns?

            include_enclosed_mass (bool) :
                Include the mass enclosed in some specified radii as one of the columns?

            include_average_quantity_inside_galaxy (bool) :
                Include the average value inside each galaxy for the quantities listed in average_quantity_data_keys?

            include_v_circ (bool) :
                Include the circular mass at some specified radii as one of the columns?

            metafile_dir (str) :
                The directory the metafiles (snapshot_times and used_parameters) are stored in.

            simulation_data_dir (str) :
                Directory containing the simulation data (used for getting the position and masses of the star particles).

            mass_radii_kwargs (dict) :
                Keyword args for self.get_mass_radii()

            enclosed_mass_ptypes (list of strs) :
                Particle types to get the mass inside a radii of.

            enclosed_mass_kwargs (dict) :
                Keyword args for self.get_enclosed_mass()

            average_quantity_data_keys (list of strs) :
                What data keys (to be passed to a standard ParticleData.get_data() function) to get the average quantity for?

            average_quantity_kwargs (dict) :
                Keyword args for self.get_average_quantity_inside_galaxy()

            v_circ_kwargs (dict) :
                Keyword args for self.get_circular_velocity()

            verbose (bool) :
                If True, print out additional information about how the steps are progressing.
        '''

        print('Saving *.AHF_halos_add for snum {}'.format( snum ))

        # Load the AHF_halos data
        self.get_halos( snum )

        # Figure out if there are any valid halos at this redshift if not, then a *lot* can be skipped.
        # TODO: Don't hard-code this in....
        valid_halos = self.halos['n_star'] >= 10
        no_valid_halos = valid_halos.sum() == 0
        blank_array = np.array( [ np.nan, ]*self.halos.index.size )

        # Create AHF_halos add
        self.halos_add = pd.DataFrame( {}, index=self.halos.index )
        self.halos_add.index.names = ['ID']

        # Get the analytic concentration
        if include_analytic_concentration:
            if verbose:
                print( "Including Analytic Concentration..." )
            self.halos_add['cAnalytic'] = self.get_analytic_concentration( metafile_dir, type_of_halo_id='halos' )

        # Get characteristic radii
        if include_mass_radii:
            if verbose:
                print( "Including Mass Radii..." )

            if no_valid_halos:
                mass_radii = [ blank_array, ]*len( mass_radii_kwargs['mass_fractions'] )
            else:
                mass_radii = self.get_mass_radii( simulation_data_dir = simulation_data_dir, **mass_radii_kwargs )

            for i, mass_fraction in enumerate( mass_radii_kwargs['mass_fractions'] ):
                label = 'Rstar{}'.format( mass_fraction )
                self.halos_add[label] = mass_radii[i]

        # Get mass enclosed in a particular radius
        if include_enclosed_mass:
            if verbose:
                print( "Including Enclosed Mass..." )
            for i, ptype in enumerate( enclosed_mass_ptypes ):

                if no_valid_halos:
                    halo_masses = blank_array
                else:
                    halo_masses = self.get_enclosed_mass( simulation_data_dir, ptype, **enclosed_mass_kwargs )

                label = self.key_parser.get_enclosed_mass_key( ptype, enclosed_mass_kwargs['galaxy_cut'], \
                                                                                                              enclosed_mass_kwargs['length_scale'], )
                self.halos_add[label] = halo_masses

        # Get average quantity inside each galaxy (for halos that have galaxies)
        if include_average_quantity_inside_galaxy:
            if verbose:
                print( "Including Average Quantities..." )

            for i, data_key in enumerate( average_quantity_data_keys ):

                if verbose:
                    print( "  Finding average {}...".format( data_key ) )

                if no_valid_halos:
                    average_quantity = blank_array
                else:
                    average_quantity = self.get_average_quantity_inside_galaxy(
                        data_key,
                        simulation_data_dir,
                        **average_quantity_inside_galaxy_kwargs
                    )

                label = self.key_parser.get_average_quantity_key(
                    data_key,
                    average_quantity_inside_galaxy_kwargs['ptype'],
                    average_quantity_inside_galaxy_kwargs['galaxy_cut'],
                    average_quantity_inside_galaxy_kwargs['length_scale'],
                )

                self.halos_add[label] = average_quantity

        # Get circular velocity at a particular radius
        if include_v_circ:
            if verbose:
                print( "Including Circular Velocity..." )
            v_circ = self.get_circular_velocity( metafile_dir=metafile_dir, **v_circ_kwargs )

            label = self.key_parser.get_velocity_at_radius_key(
                'Vc',
                v_circ_kwargs['galaxy_cut'],
                v_circ_kwargs['length_scale']
            ) 

            self.halos_add[label] = v_circ

        # Save AHF_halos add
        save_filepath = '{}_add'.format( self.halos_path )
        self.halos_add.to_csv( save_filepath, sep='\t' )

    ########################################################################

    def save_multiple_halos_adds( self, metafile_dir, snum_start, snum_end, snum_step ):
        '''Save additional columns that would be part of *.AHF_halos files, if that didn't break AHF.
        Do this for every *.AHF_halos file in self.data_dir.

        Args:
            metafile_dir (str): The directory the metafiles (snapshot_times and used_parameters) are stored in.
            snum_start (int): Starting snapshot.
            snum_end (int): Ending snapshot.
            snum_step (int): Step between snapshots.
        '''

        # Save the halos
        for snum in range( snum_start, snum_end+snum_step, snum_step):

            # Save the data
            self.save_halos_add( snum, metafile_dir )




#!/usr/bin/env python
'''Script to make movies using ffmpeg

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import os
import shutil
import sys

import galaxy_dive.plot_data.plotting as gen_plot

########################################################################
########################################################################

movie_dir_rel_path = sys.argv[1]

if len( sys.argv ) > 2:
  movie_quality = int( sys.argv[2] )
else:
  movie_quality = 1

movie_dir = os.path.abspath( movie_dir_rel_path )

dirs_to_use = os.listdir( movie_dir )

for dir_to_use in dirs_to_use:

  if os.path.isdir( dir_to_use ):

    img_dir = os.path.join( movie_dir, dir_to_use )

    file_pattern = '{}_*.png'.format( dir_to_use )

    movie_save_file = os.path.join( movie_dir, '{}.mp4'.format( dir_to_use ) )

    gen_plot.make_movie( img_dir, file_pattern, movie_save_file, movie_quality )



#!/usr/bin/env python
'''Tools for reading Rockstar output files.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import glob
import numpy as np
import os
import pandas as pd
import unyt

import galaxy_dive.read_data.metafile as read_metafile

########################################################################

default = object()

########################################################################
########################################################################

class RockstarReader( object ):
    '''Read Rockstar data.
    Note! All positions are in comoving coordinates,
    and everything has 1/h's sprinkled throughout.
    '''

    def __init__( self, sdir ):
        '''Initializes.

        Args:
            sdir (str): Simulation directory to load the Rockstar data from.
        '''

        self.data_dir = sdir

        # Common global names, and what columns those correspond to in the
        # data
        self.general_use_data_names = {
            'X' : 'X',
            'Y' : 'Y',
            'Z' : 'Z',
        }

        # Data units
        self.units = {
            'M200b' : unyt.Msun,
            'Vmax' : unyt.km / unyt.s,
            'Vmax' : unyt.km / unyt.s,
            'R200b' : unyt.Mpc,
            'Rs' : unyt.Mpc,
            'X' : unyt.Mpc,
            'Y' : unyt.Mpc,
            'Z' : unyt.Mpc,
            'VX' : unyt.km / unyt.s,
            'VY' : unyt.km / unyt.s,
            'VZ' : unyt.km / unyt.s,
            'JX' : unyt.Msun * unyt.Mpc * unyt.km / unyt.s,
            'JY' : unyt.Msun * unyt.Mpc * unyt.km / unyt.s,
            'JZ' : unyt.Msun * unyt.Mpc * unyt.km / unyt.s,
            'rs_klypin' : unyt.Mpc,
            'M200b_all' : unyt.Msun,
            'Mvir' : unyt.Msun,
            'M200c' : unyt.Msun,
            'M500c' : unyt.Msun,
            'M180b' : unyt.Msun,
        }

    ########################################################################
    # Load Data
    ########################################################################

    def get_halos( self, snum, force_reload=False ):
        '''Get out_*.list file for a particular snapshot.

        Args:
            snum (int): Snapshot number to load.

            force_reload (bool): Force reloading, even if there's already an halos file loaded.

        Modifies:
            self.halos (pd.DataFrame): Dataframe containing the requested data.
        '''

        if hasattr( self, 'halos' ):
            if (self.halos_snum == snum) and not force_reload:
                return

        # Load the data
        self.halos_path = os.path.join(
            self.data_dir,
            'out_{:03d}.list'.format( snum ),
        )
        self.halos = pd.read_csv(
            self.halos_path,
            sep = ' ',
            header = 0,
            skiprows = range(1,16),
            index_col = 0
        )

        # Rename the index to a more suitable name, without the '#'
        self.halos.index.names = ['ID']

        # Save the snapshot number of the rockstar halos file.
        self.halos_snum = snum

        # Note that we haven't added the additional halos data yet.
        self.halos_added = False




#!/usr/bin/env python
'''Tools for running executables

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import sys

########################################################################

def choose_config_or_commandline( config_vals ):
    '''This is useful for running notebooks.
    It allows the user to use commandline arguments to replace defaults
    in the notebook.

    Args:
        config_vals (list-like) :
            Default values when no commandline args are used.

    Returns
        used_vals (list-like) :
            Values used, replacing default values with provided commandline
            values.
    '''

    if sys.argv[1] != 'use_commandline':
        return config_vals

    for i in range( len( sys.argv ) - 2 ):
        config_vals[i] = type( config_vals[i] )( sys.argv[i+2] )

    return config_vals
        
    



#!/usr/bin/env python
'''Tools for generic data operations

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import copy
import numpy as np
import scipy
import scipy.signal

import galaxy_dive.utils.transformations as transformations

########################################################################
# Functions for doing simple things to data
########################################################################


def apply_floor( arr, floor ):
    '''Raises all values in the array below the floor to the floor value.

    Args:
        arr (np.arr): Array to raise the values of.
        floor (float): Floor value that will act as the minimum.

    Returns:
        floored_arr (np.arr): Array with all values below floor set to floor.
    '''

    return np.where( arr >= floor, arr, floor * np.ones(arr.shape))

########################################################################


def merge_dict( dict1, dict2 ):
    '''Merge two dictionaries, with dict2's keyword overridding dict1's keywords when there are duplicates

    Args:
        dict1 (dict): First dictionary.
        dict2 (dict): Second dictionary.

    Returns:
        merged_dict (dict): Merged dictionary.
    '''

    merged_dict = copy.deepcopy(dict1)

    merged_dict.update(dict2)

    return merged_dict

########################################################################


def block_mean( arr, fact ):
    '''Downsample an array

    Args:
        arr (np.arr): Array to downsample.
        fact (float): Factor by which to downsample.

    Returns:
        res (np.arr): Downsampled array.
    '''

    assert isinstance(fact, int), type(fact)

    sx, sy = arr.shape
    X, Y = np.ogrid[0:sx, 0:sy]
    regions = sy / fact * (X / fact) + Y / fact

    res = ndimage.mean(arr, labels=regions, index=np.arange(regions.max() + 1))
    res.shape = (sx / fact, sy / fact)

    return res

########################################################################


def cartesian( arrays, out=None ):
    """
    Generate a cartesian product of input arrays.
    Produced on Stack Overflow: http://stackoverflow.com/a/1235363/6651313

    Args:
        arrays (list of array-like): 1-D arrays to form the cartesian product of.
        out (ndarray): Array to place the cartesian product in.

    Returns:
        out (ndarray) 2-D array of shape (M, len(arrays)) containing cartesian products
                formed of input arrays.

    Examples:
        >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
        array([[1, 4, 6],
                      [1, 4, 7],
                      [1, 5, 6],
                      [1, 5, 7],
                      [2, 4, 6],
                      [2, 4, 7],
                      [2, 5, 6],
                      [2, 5, 7],
                      [3, 4, 6],
                      [3, 4, 7],
                      [3, 5, 6],
                      [3, 5, 7]])
    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
            out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n / arrays[0].size
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
            cartesian(arrays[1:], out=out[0:m, 1:])
            for j in xrange(1, arrays[0].size):
                    out[j * m:(j + 1) * m, 1:] = out[0:m, 1:]
    return out

########################################################################


def rotate_coordinates( arr, angle, axis, point=None ):
    '''Rotate a series of 3d coordinates clockwise looking down an axis.

    Args:
        arr (ndarray):
            The coordinates to be rotated. Dimensions must be (n_coords, 3).

        angle (float):
            The angle by which to rotate. Positive means clockwise.

        axis (vector arr):
            The axis to rotate around.

        point (vector arr):
            The point to rotate around. If none, defaults to the origin.

    Returns:
        rotated_arr (ndarray): The rotated array.
    '''

    # Get the rotation matrix out
    rot_matrix_4d = transformations.rotation_matrix( angle, axis, point )

    # Do only in 3d
    rot_matrix = rot_matrix_4d[:3, :3]

    # Apply the rotation
    rotated_arr = np.dot( arr, rot_matrix )

    # Set any low values to 0
    rotated_arr = np.where( np.abs( rotated_arr ) <= 1e-15, 0., rotated_arr )

    return rotated_arr

########################################################################


def align_axes(
    arr,
    new_z_axis,
    point = np.array([0., 0., 0.]),
    align_frame = True,
):
    '''Rotate a series of 3d coordinates s.t. the new z axis is aligned with
    the input vector new_z_axis

    Args:
        arr (ndarray):
            The coordinates to be rotated. Dimensions must be (n_coords, 3).

        new_z_axis (vector arr):
            The vector along which the new z axis will be aligned.

        point (vector arr):
            The point to rotate around. If none, defaults to the origin.

        align_frame (bool):
            If True, the coordinates are aligned along new_z_axis, but
            the data keep their same coordinates as before. I.e. we rotate our
            perspective s.t. new_z_axis is the new z axis.
            If False, we rotate the data instead, and keep the coordinates
            the same. I.e. we rotate the data z-axis to be aligned with
            new_z_axis, but we still view the data from the original
            perspective.
            This just changes the sign of the angle of rotation.

    Returns:
        rotated_arr (ndarray): The rotated array.
    '''
    z_axis = np.array([0., 0., 1.])

    if align_frame:
        angle_sign = 1.
    else:
        angle_sign = -1.

    # If nothing to rotate, don't try.
    if np.allclose( z_axis, new_z_axis ):
        return arr

    # Find the angle we want to rotate by
    angle = np.arccos( np.dot( z_axis, new_z_axis ) / ( np.linalg.norm( z_axis ) * np.linalg.norm( new_z_axis ) ) )
    # Based on how we define our angles, we need to multiply by 1 to rotate in
    # the opposite direction
    angle *= angle_sign

    # Find the vector we want to rotate around
    rot_vec = np.cross( z_axis, new_z_axis )

    # Rotate the array. Steps: 1. Move to the point you'll rotate around. 2. Rotate. 3. Move back.
    arr_to_be_rotated = arr - point
    rotated_arr_displaced = rotate_coordinates( arr_to_be_rotated, angle, rot_vec )
    rotated_arr = rotated_arr_displaced + point

    return rotated_arr

########################################################################


def contiguous_regions( condition ):
    '''Finds contiguous True regions of the boolean array "condition".
    Taken from https://stackoverflow.com/a/4495197

    Args:
        condition (boolean array-like):
            Input array.

    Returns:
        2D array-like:
            The first column is the start index of the region and the
            second column is the end index.
    '''

    # Find the indicies of changes in "condition"
    d = np.diff(condition)
    idx, = d.nonzero()

    # We need to start things after the change in
    # "condition". Therefore,
    # we'll shift the index by 1 to the right.
    idx += 1

    if condition[0]:
        # If the start of condition is True prepend a 0
        idx = np.r_[0, idx]

    if condition[-1]:
        # If the end of condition is True, append the length of the array
        idx = np.r_[idx, condition.size]  # Edit

    # Reshape the result into two columns
    idx.shape = (-1, 2)
    return idx

########################################################################


def cumsum2d( arr, axes=[0, 1], directions=[1, 1], ):
    '''Get the cumulative sum of an array.

    Args:
        arr (array-like) :
            Array to get the sum of.

        axes (list) :
            What axes to sum over.

        directions (list) :
            What direction to sum along.
                1 -> standard direction
                -1 -> reverse direction

    Returns:
        summed_arr (np.ndarray) :
            The i,j element of summed_arr is the sum of arr[:i+1,:j+1]
    '''

    summed_arr = np.copy( arr )
    for i, axis in enumerate( axes ):

        # Interpret directions, watching out for mistakes
        if directions[i] == 1:
            flip = False
        elif directions[i] == -1:
            flip = True
        else:
            raise Exception( "Unknown direction, {}".format( directions[i] ) )

        # Flip to change direction for summing
        if flip:
            summed_arr = np.flip( summed_arr, axis=axis )

        # Do the sum
        summed_arr = np.cumsum( summed_arr, axis=axis )

        # Flip back so that the array itself is unchanged
        if flip:
            summed_arr = np.flip( summed_arr, axis=axis )

    return summed_arr

########################################################################

def smooth_arr( arr, width=20, std=10 ):
    '''Apply a gaussian filter to an array.
    Credit to Cameron Hummels for the nice simple code I'm using.

    Args:
        arr (1D np.ndarray) :
            array to smooth

        width (int) :
            Number of elements over which to smooth.

        std (int) :
            Sigma of the guassian for which to smooth.

    Returns:
        smoothed_arr (1D np.ndarray) :
            Smoothed version of arr.
    '''

    b = scipy.signal.gaussian( width, std )
    smoothed_arr = scipy.ndimage.filters.convolve1d( arr, b/b.sum() )

    return smoothed_arr

def smooth( x, window_len=11, window='flat' ):                                    
    '''Simple moving-window signal smoothing (from scipy cookbook)

    Args:
        x (1D np.ndarray) :
            array to smooth

        wind_len (int) :
            Number of elements over which to smooth.

        window (str) :
            Type of window.

    Returns:
        smoothed_arr (1D np.ndarray) :
            Smoothed version of arr.
    '''

    if x.ndim != 1:                                                          
        raise ValueError( "smooth only accepts 1 dimension arrays." )
    if x.size < window_len:                                                  
        raise ValueError( "Input vector needs to be bigger than window size." )
    if window_len<3:                                                         
        return x                                                         
    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']: 
        raise ValueError( "Window is not one of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'" )

    s=np.r_[2*x[0]-x[window_len-1::-1],x,2*x[-1]-x[-1:-window_len:-1]]       

    if window == 'flat': #moving average                                     
        w=np.ones(window_len,'d')                                        
    else:                                                                    
        w=eval('np.'+window+'(window_len)')                              

    y=np.convolve(w/w.sum(),s,mode='same')                                   
    return y[window_len:-window_len+1]    



#!/usr/bin/env python
'''Script for generating derived data documentation.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import sys

import galaxy_dive.utils.utilities as utilities

utilities.gen_derived_data_doc( *sys.argv[1:] )
